from typing import List, Union, Optional, Tuple
import numpy as np
from scipy.interpolate import BPoly
from .interpolator import AbstractGeometricPath


class SimplePath(AbstractGeometricPath):
    """A class for representing continuous multi-dimentional function.

    This geometric path is specified by positions, velocities
    (optional) and acceleration (optional). Internally a scipy.PPoly
    instance is used to store the path. The polynomial degree depends
    on the input.

    If velocity is not given, they will be computed automatically.

    Parameters
    ------------
     x:
       "Time instances" of the waypoints.
     y:
       Function values at waypoints.
     yd:
       First-derivatives. If not given (None) will be computed
       automatically.

    """

    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        yd: Optional[np.ndarray] = None,
    ) -> None:
        if len(y.shape) == 1:
            y = y.reshape(-1, 1)
        if yd is not None and len(yd.shape) == 1:
            yd = yd.reshape(-1, 1)
        self._x = x
        self._y = y.astype(float)
        self._yd = yd if yd is None else yd.astype(float)
        self._polys = self._construct_polynomials()

    def __call__(self, xi, order=0):
        """Evaluate the path at given position."""
        ret = [poly.derivative(order)(xi) for poly in self._polys]
        return np.array(ret).T

    @property
    def dof(self) -> int:
        return self._y.shape[1]

    @property
    def path_interval(self) -> np.ndarray:
        return np.array([self._x[0], self._x[-1]], dtype=float)

    @property
    def waypoints(self) -> np.ndarray:
        return self._y

    def _autofill_yd(self):
        if self._yd is None:
            _yd = np.zeros_like(self._y[:], dtype=float)
            for i in range(1, len(_yd) - 1):
                _yd[i] = (self._y[i + 1] - self._y[i - 1]) / (
                    self._x[i + 1] - self._x[i - 1]
                )
        else:
            _yd = np.array(self._yd[:])
        return _yd

    def _construct_polynomials(self):
        polys = []
        _yd = self._autofill_yd()

        for j in range(self._y.shape[1]):

            y_with_derivatives = np.vstack((self._y[:, j], _yd[:, j])).T
            poly = BPoly.from_derivatives(self._x, y_with_derivatives)
            polys.append(poly)

        return polys



"""Parametrization algorithms
------------------------

TOPPRA
^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.algorithm.TOPPRA
   :show-inheritance:
   :members: compute_trajectory, compute_parameterization

TOPPRAsd
^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.algorithm.TOPPRAsd
   :members: compute_trajectory, set_desired_duration, compute_parameterization
   :show-inheritance:

ParameterizationReturnCode
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: toppra.algorithm.ParameterizationReturnCode
   :members:

   .. autoattribute:: Ok
   .. autoattribute:: ErrUnknown
   .. autoattribute:: ErrShortPath
   .. autoattribute:: FailUncontrollable
   .. autoattribute:: ErrForwardPassFail

ParameterizationData
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: toppra.algorithm.ParameterizationData
   :members:

[abstract]ParameterizationAlgorithm
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: toppra.algorithm.ParameterizationAlgorithm
   :members:

"""
from .algorithm import ParameterizationAlgorithm, ParameterizationData, ParameterizationReturnCode
from .reachabilitybased import TOPPRA, TOPPRAsd

__all__ = ["ParameterizationData", "ParameterizationAlgorithm", "TOPPRA", "TOPPRAsd"]



"""Internal module for exposing cpp bindings """
try:
    from toppra.cpp.toppra_int import *
    TOPPRA_INT_LOADED = True
except ImportError as err:
    print(err)
    TOPPRA_INT_LOADED = False

def bindings_loaded():
    return TOPPRA_INT_LOADED
    



"""This module defines the abstract linear constraint class."""
import numpy as np
from .constraint import ConstraintType, DiscretizationType, Constraint
from ..interpolator import AbstractGeometricPath


class LinearConstraint(Constraint):
    """A core type of constraints.

    Also known as Second-order Constraint.

    A Canonical Linear Constraint has the following form:

    .. math::

        \mathbf a_i u + \mathbf b_i x + \mathbf c_i &= v, \\\\
        \mathbf F_i v &\\leq \mathbf g_i, \\\\
        x^{bound}_{i, 0} \\leq x &\\leq x^{bound}_{i, 1}, \\\\
        u^{bound}_{i, 0} \\leq u &\\leq u^{bound}_{i, 1}.

    Alternatively, if :math:`\mathbf F_i` is constant for all values
    of :math:`i`, then we can consider the simpler constraint:

    .. math::
        \mathbf F v &\\leq \mathbf w, \\\\

    In this case, the returned value of :math:`F` by
    `compute_constraint_params` has shape (k, m) instead of (N, k, m),
    :math:`w` shape (k) instead of (N, k) and the class attribute
    `identical` will be True.

    .. note::

        Derived classes of :class:`LinearConstraint` should at
        least implement the method :func:`compute_constraint_params`.


    .. seealso::

        :class:`JointAccelerationConstraint`
        :class:`JointVelocityConstraint`
        :class:`CanonicalLinearSecondOrderConstraint`

    """
    def __init__(self):
        self.constraint_type = ConstraintType.CanonicalLinear
        self.discretization_type = DiscretizationType.Collocation
        self.n_extra_vars = 0
        self.identical = False

    def compute_constraint_params(
        self, path: AbstractGeometricPath, gridpoints: np.ndarray, *args, **kwargs
    ):
        """Compute numerical coefficients of the given constraint.

        Parameters
        ----------
        path: :class:`Interpolator`
            The geometric path.
        gridpoints: np.ndarray
            Shape (N+1,). Gridpoint use for discretizing path.

        Returns
        -------
        a: np.ndarray or None
            Shape (N + 1, m). See notes.
        b: np.ndarray, or None
            Shape (N + 1, m). See notes.
        c: np.ndarray or None
            Shape (N + 1, m). See notes.
        F: np.ndarray or None
            Shape (N + 1, k, m). See notes.
        g: np.ndarray or None
            Shape (N + 1, k,). See notes
        ubound: np.ndarray, or None
            Shape (N + 1, 2). See notes.
        xbound: np.ndarray or None
            Shape (N + 1, 2). See notes.

        """
        raise NotImplementedError


def canlinear_colloc_to_interpolate(a, b, c, F, g, xbound, ubound, gridpoints, identical=False):
    """ Convert a set of parameters to the interpolation discretization scheme.

    If a set of parameters is None, the resulting set is also None.

    Parameters
    ----------
    a: np.ndarray or None
        Shape (N + 1, m). See notes.
    b: np.ndarray or None
        Shape (N + 1, m). See notes.
    c: np.ndarray or None
        Shape (N + 1, m). See notes.
    F: np.ndarray or None
        Shape (N + 1, k, m). See notes.
    g: np.ndarray or None
        Shape (N + 1, k,). See notes
    ubound: np.ndarray, or None
        Shape (N + 1, 2). See notes.
    xbound: np.ndarray or None
        Shape (N + 1, 2). See notes.
    gridpoints: np.ndarray
        Shape (N+1,). The path discretization.
    identical: bool, optional
        If True, matrices F and g are identical at all gridpoint.

    Returns
    -------
    a_intp: np.ndarray, or None
        Shape (N + 1, m). See notes.
    b_intp: np.ndarray, or None
        Shape (N + 1, m). See notes.
    c_intp: np.ndarray, or None
        Shape (N + 1, m). See notes.
    F_intp: np.ndarray, or None
        Shape (N + 1, k, m). See notes.
    g_intp: np.ndarray, or None
        Shape (N + 1, k,). See notes
    ubound: np.ndarray, or None
        Shape (N + 1, 2). See notes.
    xbound: np.ndarray, or None
        Shape (N + 1, 2). See notes.

    """
    if a is None:
        a_intp = None
        b_intp = None
        c_intp = None
        F_intp = None
        g_intp = None
    elif not identical:
        N = a.shape[0] - 1
        d = a.shape[1]
        m = g.shape[1]
        deltas = np.diff(gridpoints)

        a_intp = np.zeros((N + 1, 2 * d))
        a_intp[:, :d] = a
        a_intp[:-1, d:] = a[1:] + 2 * deltas.reshape(-1, 1) * b[1:]
        a_intp[-1, d:] = a_intp[-1, :d]

        b_intp = np.zeros((N + 1, 2 * d))
        b_intp[:, :d] = b
        b_intp[:-1, d:] = b[1:]
        b_intp[-1, d:] = b_intp[-1, :d]

        c_intp = np.zeros((N + 1, 2 * d))
        c_intp[:, :d] = c
        c_intp[:-1, d:] = c[1:]
        c_intp[-1, d:] = c_intp[-1, :d]

        g_intp = np.zeros((N + 1, 2 * m))
        g_intp[:, :m] = g
        g_intp[:-1, m:] = g[1:]
        g_intp[-1, m:] = g_intp[-1, :m]

        F_intp = np.zeros((N + 1, 2 * m, 2 * d))
        F_intp[:, :m, :d] = F
        F_intp[:-1, m:, d:] = F[1:]
        F_intp[-1, m:, d:] = F[-1]
    elif identical:
        N = a.shape[0] - 1
        m, d = F.shape
        deltas = np.diff(gridpoints)

        a_intp = np.zeros((N + 1, 2 * d))
        a_intp[:, :d] = a
        a_intp[:-1, d:] = a[1:] + 2 * deltas.reshape(-1, 1) * b[1:]
        a_intp[-1, d:] = a_intp[-1, :d]

        b_intp = np.zeros((N + 1, 2 * d))
        b_intp[:, :d] = b
        b_intp[:-1, d:] = b[1:]
        b_intp[-1, d:] = b_intp[-1, :d]

        c_intp = np.zeros((N + 1, 2 * d))
        c_intp[:, :d] = c
        c_intp[:-1, d:] = c[1:]
        c_intp[-1, d:] = c_intp[-1, :d]

        g_intp = np.zeros(2 * m)
        g_intp[:m] = g
        g_intp[m:] = g

        F_intp = np.zeros((2 * m, 2 * d))
        F_intp[:m, :d] = F
        F_intp[m:, d:] = F

    return a_intp, b_intp, c_intp, F_intp, g_intp, xbound, ubound




"""Exceptions used in the toppra library."""


class ToppraError(Exception):
    """A generic error class used in the toppra library."""


class BadInputVelocities(ToppraError):
    """Raise when given input velocity is invalid."""


class SolverNotFound(ToppraError):
    """Unable to find a solver."""



from .solverwrapper import SolverWrapper
import numpy as np
from ..constraint import ConstraintType
from ..constants import INFTY

try:
    from qpoases import (
        PyOptions as Options,
        PyPrintLevel as PrintLevel,
        PyReturnValue as ReturnValue,
        PySQProblem as SQProblem,
    )

    qpoases_FOUND = True
except ImportError:
    qpoases_FOUND = False

INF = INFTY


class qpOASESSolverWrapper(SolverWrapper):
    """ A solver wrapper using `qpOASES`.

    Parameters
    ----------
    constraint_list: list of :class:`.Constraint`
        The constraints the robot is subjected to.
    path: :class:`.Interpolator`
        The geometric path.
    path_discretization: array
        The discretized path positions.
    """

    def __init__(self, constraint_list, path, path_discretization):
        assert qpoases_FOUND, "toppra is unable to find any installation of qpoases!"
        super(qpOASESSolverWrapper, self).__init__(
            constraint_list, path, path_discretization
        )

        # Currently only support Canonical Linear Constraint
        self.nC = 2  # First constraint is x + 2 D u <= xnext_max, second is xnext_min <= x + 2D u
        for i, constraint in enumerate(constraint_list):
            if constraint.get_constraint_type() != ConstraintType.CanonicalLinear:
                raise NotImplementedError
            a, b, c, F, v, _, _ = self.params[i]
            if a is not None:
                if constraint.identical:
                    self.nC += F.shape[0]
                else:
                    self.nC += F.shape[1]

        self._A = np.zeros((self.nC, self.nV))
        self._lA = -np.ones(self.nC)
        self._hA = -np.ones(self.nC)

        option = Options()
        option.printLevel = PrintLevel.NONE
        self.solver = SQProblem(self.nV, self.nC)
        self.solver.setOptions(option)

    def solve_stagewise_optim(self, i, H, g, x_min, x_max, x_next_min, x_next_max):
        # NOTE: qpOASES solve QPs of the following form:
        #  min    0.5 y^T H y + g^T y
        #  s.t    lA <= Ay <= hA
        #         l  <=  y <= h
        assert i <= self.N and 0 <= i

        l = -np.ones(2) * INF
        h = np.ones(2) * INF

        if x_min is not None:
            l[1] = max(l[1], x_min)
        if x_max is not None:
            h[1] = min(h[1], x_max)

        if i < self.N:
            delta = self.get_deltas()[i]
            if x_next_min is not None:
                self._A[0] = [-2 * delta, -1]
                self._hA[0] = -x_next_min
                self._lA[0] = -INF
            else:
                self._A[0] = [0, 0]
                self._hA[0] = INF
                self._lA[0] = -INF
            if x_next_max is not None:
                self._A[1] = [2 * delta, 1]
                self._hA[1] = x_next_max
                self._lA[1] = -INF
            else:
                self._A[1] = [0, 0]
                self._hA[1] = INF
                self._lA[1] = -INF
        cur_index = 2
        for j in range(len(self.constraints)):
            a, b, c, F, v, ubound, xbound = self.params[j]

            # Case 1
            if a is not None:
                if self.constraints[j].identical:
                    nC_ = F.shape[0]
                    self._A[cur_index : cur_index + nC_, 0] = F.dot(a[i])
                    self._A[cur_index : cur_index + nC_, 1] = F.dot(b[i])
                    self._hA[cur_index : cur_index + nC_] = v - F.dot(c[i])
                    self._lA[cur_index : cur_index + nC_] = -INF
                else:
                    nC_ = F[i].shape[0]
                    self._A[cur_index : cur_index + nC_, 0] = F[i].dot(a[i])
                    self._A[cur_index : cur_index + nC_, 1] = F[i].dot(b[i])
                    self._hA[cur_index : cur_index + nC_] = v[i] - F[i].dot(c[i])
                    self._lA[cur_index : cur_index + nC_] = -INF
                cur_index = cur_index + nC_

            if ubound is not None:
                l[0] = max(l[0], ubound[i, 0])
                h[0] = min(h[0], ubound[i, 1])

            if xbound is not None:
                l[1] = max(l[1], xbound[i, 0])
                h[1] = min(h[1], xbound[i, 1])

        if H is None:
            H = np.zeros((self.get_no_vars(), self.get_no_vars()))

        res = self.solver.init(
            H, g, self._A, l, h, self._lA, self._hA, np.array([1000])
        )
        if res == ReturnValue.SUCCESSFUL_RETURN:
            var = np.zeros(self.nV)
            self.solver.getPrimalSolution(var)
            return var
        else:
            res = np.empty(self.get_no_vars())
            res[:] = np.nan
            return res



"""Base class for all path parametrization contraints. """
from enum import Enum
import logging
import numpy as np
from ..interpolator import AbstractGeometricPath

logger = logging.getLogger(__name__)


class ConstraintType(Enum):
    """Type of path parametrization constraint."""

    #: Unknown
    Unknown = -1
    #: Simple linear constraints with only linear inequalities
    CanonicalLinear = 0
    #: Linear constraints with linear conic inequalities.
    CanonicalConic = 1


class DiscretizationType(Enum):
    """Enum to mark different Discretization Scheme for constraint.

    In general, the difference in speed is not too large. Should use
    Interpolation if possible.
    """
    #: Smaller problem size, but lower accuracy.
    Collocation = 0

    #: Larger problem size, but higher accuracy.
    Interpolation = 1


class Constraint(object):
    """The base constraint class."""

    def __repr__(self):
        string = self.__class__.__name__ + "(\n"
        string += "    Type: {:}".format(self.constraint_type) + "\n"
        string += (
            "    Discretization Scheme: {:}".format(self.discretization_type) + "\n"
        )
        string += self._format_string
        string += ")"
        return string

    def get_dof(self):
        """Return the degree of freedom of the constraint.

        TODO: It is unclear what is a dof of a constraint. Perharps remove this.
        """
        return self.dof

    def get_no_extra_vars(self):
        """Return the number of extra variable required.

        TODO: This is not a property of a constraint. Rather it is specific to kinds of constraints. To be removed.
        """
        return self.n_extra_vars

    def get_constraint_type(self):
        """Return the constraint type.

        TODO: Use property instead.

        """
        return self.constraint_type

    def get_discretization_type(self):
        """Return the discretization type.

        TODO: Use property instead.

        """
        return self.discretization_type

    def set_discretization_type(self, discretization_type):
        """Discretization type: Collocation or Interpolation.

        Parameters
        ----------
        discretization_type: int or :class:`DiscretizationType`
            Method to discretize this constraint.
        """
        if discretization_type == 0:
            self.discretization_type = DiscretizationType.Collocation
        elif discretization_type == 1:
            self.discretization_type = DiscretizationType.Interpolation
        elif (
            discretization_type == DiscretizationType.Collocation
            or discretization_type == DiscretizationType.Interpolation
        ):
            self.discretization_type = discretization_type
        else:
            raise "Discretization type: {:} not implemented!".format(
                discretization_type
            )

    def compute_constraint_params(
        self, path: AbstractGeometricPath, gridpoints: np.ndarray, *args, **kwargs
    ):
        """Evaluate parameters of the constraint."""
        raise NotImplementedError



from .solverwrapper import SolverWrapper
import logging
import numpy as np
from ..constraint import ConstraintType
from ..constants import CVXPY_MAXX, CVXPY_MAXU


logger = logging.getLogger(__name__)
try:
    import cvxpy

    FOUND_CVXPY = True
except ImportError:
    logger.info("CVXPY installation not found.")
    FOUND_CVXPY = False
try:
    import mosek

    FOUND_MOSEK = True
except ImportError:
    logger.info("Mosek installation not found!")
    FOUND_MOSEK = False


class cvxpyWrapper(SolverWrapper):
    """A solver wrapper using `cvxpy`.

    NOTE: the two constants CVXPY_MAXX and CVXPY_MAXU is used to
    guarantee that the solution is not too large, in which case cvxpy
    can't handle very well.

    `cvxpyWrapper` should not be used in production due to robustness
    issue.

    Parameters
    ----------
    constraint_list: list of :class:`.Constraint`
        The constraints the robot is subjected to.
    path: :class:`.Interpolator`
        The geometric path.
    path_discretization: array
        The discretized path positions.

    """

    def __init__(self, constraint_list, path, path_discretization):
        super(cvxpyWrapper, self).__init__(constraint_list, path, path_discretization)
        valid_types = [ConstraintType.CanonicalLinear, ConstraintType.CanonicalConic]
        # Currently only support Canonical Linear Constraint
        for constraint in constraint_list:
            if constraint.get_constraint_type() not in valid_types:
                raise NotImplementedError

    def solve_stagewise_optim(self, i, H, g, x_min, x_max, x_next_min, x_next_max):
        assert i <= self.N and 0 <= i

        ux = cvxpy.Variable(2)
        u = ux[0]
        x = ux[1]
        cvxpy_constraints = [-CVXPY_MAXU <= u, u <= CVXPY_MAXU, 0 <= x, x <= CVXPY_MAXX]

        if not np.isnan(x_min):
            cvxpy_constraints.append(x_min <= x)
        if not np.isnan(x_max):
            cvxpy_constraints.append(x <= x_max)

        if i < self.N:
            delta = self.get_deltas()[i]
            if not np.isnan(x_next_min):
                cvxpy_constraints.append(x_next_min <= x + 2 * delta * u)
            if not np.isnan(x_next_max):
                cvxpy_constraints.append(x + 2 * delta * u <= x_next_max)

        for k, constraint in enumerate(self.constraints):
            if constraint.get_constraint_type() == ConstraintType.CanonicalLinear:
                a, b, c, F, h, ubound, xbound = self.params[k]

                if a is not None:
                    v = a[i] * u + b[i] * x + c[i]
                    if constraint.identical:
                        cvxpy_constraints.append(F * v <= h)
                    else:
                        cvxpy_constraints.append(F[i] * v <= h[i])

                # ecos (via cvxpy in this class) is very bad at
                # handling badly scaled problems. Problems with very
                # large bound. The below max(), min() operators is a
                # workaround to get pass this issue.
                if ubound is not None:
                    cvxpy_constraints.append(max(-CVXPY_MAXU, ubound[i, 0]) <= u)
                    cvxpy_constraints.append(u <= min(CVXPY_MAXU, ubound[i, 1]))

                if xbound is not None:
                    cvxpy_constraints.append(xbound[i, 0] <= x)
                    cvxpy_constraints.append(x <= min(CVXPY_MAXX, xbound[i, 1]))

            elif constraint.get_constraint_type() == ConstraintType.CanonicalConic:
                a, b, c, P, ubound, xbound = self.params[k]

                if a is not None:
                    d = a.shape[1]
                    for j in range(d):
                        cvxpy_constraints.append(
                            a[i, j] * u
                            + b[i, j] * x
                            + c[i, j]
                            + cvxpy.norm(P[i, j].T[:, :2] * ux + P[i, j].T[:, 2])
                            <= 0
                        )

                if ubound is not None:
                    cvxpy_constraints.append(max(-CVXPY_MAXU, ubound[i, 0]) <= u)
                    cvxpy_constraints.append(u <= min(CVXPY_MAXU, ubound[i, 1]))

                if xbound is not None:
                    cvxpy_constraints.append(xbound[i, 0] <= x)
                    cvxpy_constraints.append(x <= min(CVXPY_MAXX, xbound[i, 1]))

        if H is None:
            H = np.zeros((self.get_no_vars(), self.get_no_vars()))

        objective = cvxpy.Minimize(0.5 * cvxpy.quad_form(ux, H) + g * ux)
        problem = cvxpy.Problem(objective, constraints=cvxpy_constraints)
        try:
            problem.solve(verbose=False)
        except cvxpy.SolverError:
            # solve fail
            pass
        if (
            problem.status == cvxpy.OPTIMAL
            or problem.status == cvxpy.OPTIMAL_INACCURATE
        ):
            return np.array(ux.value).flatten()
        else:
            res = np.empty(self.get_no_vars())
            res[:] = np.nan
            return res



from .interpolator import RaveTrajectoryWrapper, SplineInterpolator
from .constraint import (
    JointAccelerationConstraint,
    JointVelocityConstraint,
    DiscretizationType,
    SecondOrderConstraint,
)
from .algorithm import TOPPRA
import numpy as np
import logging
import time

logger = logging.getLogger(__name__)


def retime_active_joints_kinematics(
    traj,
    robot,
    output_interpolator=False,
    vmult=1.0,
    amult=1.0,
    N=100,
    use_ravewrapper=False,
    additional_constraints=[],
    solver_wrapper="hotqpoases",
):
    """ Retime a trajectory wrt velocity and acceleration constraints.

    Parameters
    ----------
    traj: OpenRAVE Trajectory or (N,dof)array
        The original trajectory. If is an array, a cubic spline will be used to interpolate
        through all points before parameterization.
    robot: OpenRAVE Robot
        The kinematic limits, which are velocity and acceleration limits, are taken from the robot
        model.
    output_interpolator: bool, optional
        If this is true, output the interpolator object together with the retimed trajectory.
    vmult: float, optional
        Safety factor.
    amult: float, optional
        Safety factor.
    N: int, optional
        Approximate number of gridpoints.
    use_ravewrapper: bool, optional
        If is true, use the input openrave trajectory directly for parameterization. This is,
        however, not desirable because the input trajectory tends to have discontinous second
        order derivative. This causes alots of problems for the parameterization algorithm
        TOPPRA.
    additional_constraints: list, optional
        List of additional constraints to consider.

    Returns
    -------
    traj_rave: OpenRAVE Trajectory
        The retimed trajectory. Return None if retiming fails.
    traj_ra: SplineInterpolator
        Return if 'output_interpolator' is True.
    """
    _t0 = time.time()
    if isinstance(traj, np.ndarray):
        logger.debug("Received a list of waypoints.")
        ss_waypoints = np.linspace(0, 1, len(traj))
        path = SplineInterpolator(ss_waypoints, traj, bc_type="natural")
    elif use_ravewrapper:
        logger.warning("Use RaveTrajectoryWrapper. This might not work properly!")
        path = RaveTrajectoryWrapper(traj, robot)
    elif isinstance(traj, SplineInterpolator):
        path = traj
    else:
        logger.debug("Use a spline to represent the input path!")
        ss_waypoints = []
        waypoints = []
        spec = traj.GetConfigurationSpecification()
        for i in range(traj.GetNumWaypoints()):
            data = traj.GetWaypoint(i)
            dt = spec.ExtractDeltaTime(data)
            # If delta is too small, skip it.
            if dt > 1e-5 or len(waypoints) == 0:
                if len(ss_waypoints) == 0:
                    ss_waypoints.append(0)
                else:
                    ss_waypoints.append(ss_waypoints[-1] + dt)
                waypoints.append(
                    spec.ExtractJointValues(data, robot, robot.GetActiveDOFIndices())
                )
        path = SplineInterpolator(ss_waypoints, waypoints)

    vmax = robot.GetActiveDOFMaxVel() * vmult
    amax = robot.GetActiveDOFMaxAccel() * amult
    vlim = np.vstack((-vmax, vmax)).T
    alim = np.vstack((-amax, amax)).T

    pc_vel = JointVelocityConstraint(vlim)
    pc_acc = JointAccelerationConstraint(
        alim, discretization_scheme=DiscretizationType.Interpolation
    )
    logger.debug("Number of constraints %d", 2 + len(additional_constraints))
    logger.debug(str(pc_vel))
    logger.debug(str(pc_acc))
    for _c in additional_constraints:
        logger.debug(str(_c))

    # Include the waypoints in the grid
    ds = path.ss_waypoints[-1] / N
    gridpoints = [path.ss_waypoints[0]]
    for i in range(len(path.ss_waypoints) - 1):
        Ni = int(np.ceil((path.ss_waypoints[i + 1] - path.ss_waypoints[i]) / ds))
        gridpoints.extend(
            path.ss_waypoints[i]
            + np.linspace(0, 1, Ni + 1)[1:]
            * (path.ss_waypoints[i + 1] - path.ss_waypoints[i])
        )
    instance = TOPPRA(
        [pc_vel, pc_acc] + additional_constraints,
        path,
        gridpoints=gridpoints,
        solver_wrapper=solver_wrapper,
    )
    _t1 = time.time()

    traj_ra = instance.compute_trajectory(0, 0)
    _t2 = time.time()
    logger.debug(
        "t_setup={:.5f}ms, t_solve={:.5f}ms, t_total={:.5f}ms".format(
            (_t1 - _t0) * 1e3, (_t2 - _t1) * 1e3, (_t2 - _t0) * 1e3
        )
    )
    if traj_ra is None:
        logger.warning("Retime fails.")
        traj_rave = None
    else:
        logger.debug("Retime successes!")
        traj_rave = traj_ra.compute_rave_trajectory(robot)

    if output_interpolator:
        return traj_rave, traj_ra
    else:
        return traj_rave


def create_rave_torque_path_constraint(
    robot, discretization_scheme=DiscretizationType.Collocation
):
    """Create torque bound for the given robot.

    The torque bound constraint for a manipulator whose links are
    rigid bodies is a Second Order canonical linear constraint.

    Parameters
    ----------
    robot: openravepy.Robot


    Returns
    -------
    cnst: `SecondOrderCanonicalLinearConstraint`
    """
    qdd_full = np.zeros(robot.GetDOF())
    active_dofs = robot.GetActiveDOFIndices()

    def inv_dyn(q, qd, qdd, s=0):
        with robot:
            # Temporary remove vel/acc constraints
            vlim = robot.GetDOFVelocityLimits()
            alim = robot.GetDOFAccelerationLimits()
            robot.SetDOFVelocityLimits(100 * vlim)
            robot.SetDOFAccelerationLimits(100 * alim)
            # Inverse dynamics
            qdd_full[active_dofs] = qdd
            robot.SetActiveDOFValues(q)
            robot.SetActiveDOFVelocities(qd)
            res = robot.ComputeInverseDynamics(qdd_full)
            # Restore vel/acc constraints
            robot.SetDOFVelocityLimits(vlim)
            robot.SetDOFAccelerationLimits(alim)
        return res[active_dofs]

    tau_max = robot.GetDOFTorqueLimits()[robot.GetActiveDOFIndices()]
    F = np.vstack((np.eye(robot.GetActiveDOF()), -np.eye(robot.GetActiveDOF())))
    g = np.hstack((tau_max, tau_max))

    def cnst_F(q):
        return F

    def cnst_g(q):
        return g

    cnst = SecondOrderConstraint(
        inv_dyn,
        cnst_F,
        cnst_g,
        dof=robot.GetActiveDOF(),
        discretization_scheme=discretization_scheme,
    )
    return cnst



from ..algorithm import ParameterizationAlgorithm, ParameterizationReturnCode
from ...constants import LARGE, SMALL, TINY, INFTY, CVXPY_MAXX, MAX_TRIES
from ...constraint import ConstraintType
import toppra.solverwrapper
import toppra.exceptions as exceptions
import toppra.interpolator as interpolator

import numpy as np
import logging

logger = logging.getLogger(__name__)


class ReachabilityAlgorithm(ParameterizationAlgorithm):
    """Base class for Reachability Analysis-based path parameterization algorithms.

    Parameters
    ----------
    constraint_list: List[:class:`~toppra.constraint.Constraint`]
        List of constraints on the robot dynamics.
    path: Interpolator
    gridpoints: np.ndarray, optional
        Shape (N+1,). Gridpoints for discretization of the path position.
    solver_wrapper: str, optional
        Name of solver to use. If is None, select the most suitable wrapper.
    parametrizer: str, optional
        Name of the output parametrizer to use.

    Notes
    -----
    There are two patterns this class implements.

    1) Together with `SolverWrapper`, it uses a Strategy pattern to achieve different
       solver configuration.

    2) The class itself implements a `Template` pattern.

    RA-based algorithm uses a `SolverWrapper` for most, if not all, computations.
    During initialization, a solver wrapper is constructed from the given variables
    and is used afterward.

    In addition to a generic path parameterization algorithm, a RA-based algorithm
    implement additionally three methods:
    - compute_controllable_sets
    - compute_reachable_sets
    - compute_feasible_sets
    """

    def __init__(
            self, constraint_list, path, gridpoints=None, solver_wrapper=None, parametrizer=None, **kwargs
    ):
        super(ReachabilityAlgorithm, self).__init__(
            constraint_list, path, gridpoints=gridpoints, parametrizer=parametrizer, **kwargs
        )

        # Check for conic constraints
        has_conic = False
        for c in constraint_list:
            if c.get_constraint_type() == ConstraintType.CanonicalConic:
                has_conic = True

        # Select solver wrapper automatically
        available_solvers = toppra.solverwrapper.available_solvers(output_msg=False)
        if solver_wrapper is None:
            logger.info(
                "Solver wrapper not supplied. Choose solver wrapper automatically!"
            )
            if has_conic:
                if not available_solvers["ecos"]:
                    raise exceptions.ToppraError("Solverwrapper not available.")
                solver_wrapper = "ecos"
            else:
                valid_solver = [solver for solver, avail in available_solvers if avail]
                solver_wrapper = valid_solver[0]
            logger.info("Select solver {:}".format(solver_wrapper))

        # Check solver-wrapper suitability
        if has_conic:
            assert solver_wrapper.lower() in [
                "cvxpy",
                "ecos",
            ], "Problem has conic constraints, solver {:} is not suitable".format(
                solver_wrapper
            )
        else:
            assert solver_wrapper.lower() in [
                "cvxpy",
                "qpoases",
                "ecos",
                "hotqpoases",
                "seidel",
            ], "Solver {:} not found".format(solver_wrapper)

        if solver_wrapper.lower() == "cvxpy":
            from toppra.solverwrapper.cvxpy_solverwrapper import cvxpyWrapper

            self.solver_wrapper = cvxpyWrapper(
                self.constraints, self.path, self.gridpoints
            )
        elif solver_wrapper.lower() == "qpoases":
            from toppra.solverwrapper.qpoases_solverwrapper import qpOASESSolverWrapper

            self.solver_wrapper = qpOASESSolverWrapper(
                self.constraints, self.path, self.gridpoints
            )
        elif solver_wrapper.lower() == "hotqpoases":
            from toppra.solverwrapper.hot_qpoases_solverwrapper import (
                hotqpOASESSolverWrapper,
            )

            self.solver_wrapper = hotqpOASESSolverWrapper(
                self.constraints, self.path, self.gridpoints
            )
        elif solver_wrapper.lower() == "ecos":
            from toppra.solverwrapper.ecos_solverwrapper import ecosWrapper

            self.solver_wrapper = ecosWrapper(
                self.constraints, self.path, self.gridpoints
            )
        elif solver_wrapper.lower() == "seidel":
            from toppra.solverwrapper.cy_seidel_solverwrapper import seidelWrapper

            self.solver_wrapper = seidelWrapper(
                self.constraints, self.path, self.gridpoints, solve_lp1d=True
            )
        else:
            raise NotImplementedError(
                "Solver wrapper {:} not found!".format(solver_wrapper)
            )

    def compute_feasible_sets(self):
        """Compute the sets of feasible squared velocities.

        Returns
        -------
        X: array
            Shape (N+1,2). X[i] contains the lower and upper bound of
            the feasible squared path velocity at s[i].  If there is
            no feasible state, X[i] equals (np.nan, np.nan).

        """
        logger.debug("Start computing the feasible sets")
        nV = self.solver_wrapper.get_no_vars()
        Hzero = np.zeros((nV, nV))
        g_lower = np.zeros(nV)
        g_lower[0] = 1e-9
        g_lower[1] = 1
        X = np.zeros((self._N + 1, 2))
        self.solver_wrapper.setup_solver()
        for i in range(self._N + 1):
            X[i, 0] = self.solver_wrapper.solve_stagewise_optim(
                i, Hzero, g_lower, -CVXPY_MAXX, CVXPY_MAXX, -CVXPY_MAXX, CVXPY_MAXX
            )[1]
            X[i, 1] = self.solver_wrapper.solve_stagewise_optim(
                i, Hzero, -g_lower, -CVXPY_MAXX, CVXPY_MAXX, -CVXPY_MAXX, CVXPY_MAXX
            )[1]
            if logger.getEffectiveLevel() == logging.DEBUG:
                logger.debug("X[{:d}]={:}".format(i, X[i]))
        self.solver_wrapper.close_solver()
        for i in range(self._N + 1):
            if X[i, 0] < 0:
                X[i, 0] = 0
        self._problem_data.X = X
        return X

    def compute_controllable_sets(self, sdmin, sdmax):
        """Compute the sets of controllable squared path velocities.

        Parameters
        ----------
        sdmin: float
            Lower bound on the final path velocity.
        sdmax: float
            Upper bound on the final path velocity.

        Returns
        -------
        K : array
            Shape (N+1, 2). Element K[i] contains the squared upper
            and lower controllable velocities at position s[i].

        """
        assert sdmin <= sdmax and 0 <= sdmin
        K = np.zeros((self._N + 1, 2))
        K[self._N] = [sdmin ** 2, sdmax ** 2]
        logger.debug("Start computing the controllable sets")
        self.solver_wrapper.setup_solver()
        for i in range(self._N - 1, -1, -1):
            K[i] = self._one_step(i, K[i + 1])
            if K[i, 0] < 0:
                K[i, 0] = 0
            if np.isnan(K[i]).any():
                logger.warning(
                    "A numerical error occurs: The controllable set at step "
                    "[{:d} / {:d}] can't be computed.".format(i, self._N + 1)
                )
                return K
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("[Compute controllable sets] K_{:d}={:}".format(i, K[i]))

        self.solver_wrapper.close_solver()
        return K

    def _one_step(self, i, K_next):
        """Perform the one-step operation.

        Parameters
        ----------
        i: int
            Stage index.
        K_next: array
            Shape(2,).Two ends of the set of controllable path
            velocities at stage (i+1).

        Returns
        -------
        res: array
            Shape (2,). Set of controllable squared path velocities
            K[i].

        """
        res = np.zeros(2)
        if np.isnan(K_next).any() or i < 0 or i > self._N:
            res[:] = np.nan
            return res

        nV = self.solver_wrapper.get_no_vars()
        g_upper = np.zeros(nV)
        g_upper[0] = 1e-9
        g_upper[1] = -1
        x_upper = self.solver_wrapper.solve_stagewise_optim(
            i, None, g_upper, np.nan, np.nan, K_next[0], K_next[1]
        )[1]
        x_lower = self.solver_wrapper.solve_stagewise_optim(
            i, None, -g_upper, np.nan, np.nan, K_next[0], K_next[1]
        )[1]
        res[:] = [x_lower, x_upper]
        return res

    def compute_parameterization(self, sd_start, sd_end, return_data=False):
        """Compute a path parameterization.

        If fail, whether because there is no valid parameterization or
        because of numerical error, the arrays returns should contain
        np.nan.

        Parameters
        ----------
        sd_start: float
            Starting path velocity. Must be positive.
        sd_end: float
            Goal path velocity. Must be positive.
        return_data: bool, optional
            If is True, also return matrix K which contains the controllable sets.

        Returns
        -------
        sdd_vec: array
            Shape (N,). Path accelerations. Double array. Will contain
            nan(s) if failed.

        sd_vec: array
            Shape (N+1,). Path velocities. Double array. Will contain nan(s) if failed.

        v_vec: array or None
            Shape (N,). Auxiliary variables.

        K: array
            Shape (N+1, 2). Return the controllable set if
            `return_data` is True.
        """
        if sd_end < 0 or sd_start < 0:
            raise exceptions.BadInputVelocities(
                "Negative path velocities: path velocities must be positive: (%s, %s)"
                % (sd_start, sd_end)
            )
        K = self.compute_controllable_sets(sd_end, sd_end)
        if np.isnan(K).any():
            logger.warning(
                "An error occurred when computing controllable velocities. "
                "The path is not controllable, or is badly conditioned."
            )
            self._problem_data.return_code = ParameterizationReturnCode.FailUncontrollable
            if return_data:
                return None, None, None, K
            else:
                return None, None, None
        self._problem_data.K = K

        x_start = sd_start ** 2
        if x_start + SMALL < K[0, 0] or K[0, 1] + SMALL < x_start:
            logger.warning(
                "The initial velocity is not controllable. {:f} not in ({:f}, {:f})".format(
                    x_start, K[0, 0], K[0, 1]
                )
            )
            self._problem_data.return_code = ParameterizationReturnCode.FailUncontrollable
            if return_data:
                return None, None, None, K
            else:
                return None, None, None

        N = self.solver_wrapper.get_no_stages()
        deltas = self.solver_wrapper.get_deltas()
        xs = np.zeros(N + 1)
        xs[0] = x_start
        us = np.zeros(N)
        v_vec = np.zeros((N, self.solver_wrapper.get_no_vars() - 2))

        tries = 0
        self.solver_wrapper.setup_solver()
        i = 0
        while i < self._N:
            optim_res = self._forward_step(i, xs[i], K[i + 1])
            if np.isnan(optim_res[0]):
                # NOTE: This case happens because the constraint
                # K[i + 1, 0] <= x[i] + 2D u[i] <= K[i + 1, 1]
                # become just slightly infeasible.  This happens more often
                # with ECOS, which is an interior point solver, than
                # with qpoases or seidel, both of which are active set
                # solvers. The strategy used to handle this case is in the
                # next line: simply reduce the current velocity by 0.1% or
                # a very small value (TINY) and choose the large value.
                if tries < MAX_TRIES:
                    xs[i] = max(
                        xs[i] - TINY, 0.999 * xs[i]
                    )  # a slightly more aggressive reduction
                    tries += 1
                    logger.warning(
                        "A numerical error occurs: the instance is controllable "
                        "but forward pass fails. Attempt to try again with x[i] "
                        "slightly reduced.\n"
                        "x[{:d}] reduced from {:.6f} to {:.6f}".format(
                            i, xs[i] + SMALL, xs[i]
                        )
                    )
                else:
                    logger.warning(
                        "Number of trials (to reduce xs[i]) reaches limits. "
                        "Compute parametrization fails!"
                    )
                    xs[i + 1 :] = np.nan
                    break
            else:
                tries = 0
                us[i] = optim_res[0]
                # The below function min( , max( ,)) ensure that the
                # state x_{i+1} is controllable.  While this is
                # ensured theoretically by the existence of the
                # controllable sets, numerical errors might violate
                # this condition.
                x_next = xs[i] + 2 * deltas[i] * us[i]
                x_next = max(x_next - TINY, 0.9999 * x_next)
                xs[i + 1] = min(K[i + 1, 1], max(K[i + 1, 0], x_next))
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(
                        "[Forward pass] u[{:d}] = {:f}, x[{:d}] = {:f}".format(
                            i, us[i], i + 1, xs[i + 1]
                        )
                    )
                v_vec[i] = optim_res[2:]
                i += 1
        self.solver_wrapper.close_solver()

        sd_vec = np.sqrt(xs)
        sdd_vec = np.copy(us)
        self._problem_data.sd_vec = sd_vec
        self._problem_data.sdd_vec = sdd_vec
        if np.isnan(sd_vec).any():
            self._problem_data.return_code = ParameterizationReturnCode.ErrUnknown
        else:
            self._problem_data.return_code = ParameterizationReturnCode.Ok
        if return_data:
            return sdd_vec, sd_vec, v_vec, K
        else:
            return sdd_vec, sd_vec, v_vec

    def _one_step_forward(self, i, L_current, feasible_set_next):
        res = np.zeros(2)
        if np.isnan(L_current).any() or i < 0 or i > self._N:
            res[:] = np.nan
            return res
        nV = self.solver_wrapper.get_no_vars()
        g_upper = np.zeros(nV)
        deltas = self.solver_wrapper.get_deltas()[i - 1]
        g_upper[0] = -2 * deltas
        g_upper[1] = -1

        x_next_min = feasible_set_next[0]
        x_next_max = feasible_set_next[1]

        opt_1 = self.solver_wrapper.solve_stagewise_optim(
            i, None, g_upper, L_current[0], L_current[1], x_next_min, x_next_max
        )
        x_opt_1 = opt_1[1]
        u_opt_1 = opt_1[0]
        x_upper = x_opt_1 + 2 * deltas * u_opt_1

        opt_0 = self.solver_wrapper.solve_stagewise_optim(
            i, None, -g_upper, L_current[0], L_current[1], x_next_min, x_next_max
        )
        x_opt_0 = opt_0[1]
        u_opt_0 = opt_0[0]
        x_lower = x_opt_0 + 2 * deltas * u_opt_0

        res[:] = [x_lower, x_upper]
        return res

    def compute_reachable_sets(self, sdmin, sdmax):
        assert sdmin <= sdmax and 0 <= sdmin
        feasible_sets = self.compute_feasible_sets()
        L = np.zeros((self._N + 1, 2))
        L[0] = [sdmin ** 2, sdmax ** 2]
        logger.debug("Start computing the reachable sets")
        self.solver_wrapper.setup_solver()
        for i in range(0, self._N):
            L[i + 1] = self._one_step_forward(i, L[i], feasible_sets[i + 1])
            if L[i + 1, 0] < 0:
                L[i + 1, 0] = 0
            if np.isnan(L[i + 1]).any():
                logger.warn(
                    "L[{:d}]={:}. Path not parametrizable.".format(i + 1, L[i + 1])
                )
                return L
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    "[Compute reachable sets] L_{:d}={:}".format(i + 1, L[i + 1])
                )

        self.solver_wrapper.close_solver()
        return L



from .solverwrapper import SolverWrapper
import numpy as np
from ..constraint import ConstraintType
from ..constants import QPOASES_INFTY, TINY, SMALL
from ..exceptions import SolverNotFound


try:
    from qpoases import (
        PyOptions as Options,
        PyPrintLevel as PrintLevel,
        PyReturnValue as ReturnValue,
        PySQProblem as SQProblem,
    )

    qpoases_FOUND = True
except ImportError:
    qpoases_FOUND = False
import logging

logger = logging.getLogger(__name__)


class hotqpOASESSolverWrapper(SolverWrapper):
    """`qpOASES` solver wrapper with hot-start.

    This wrapper takes advantage of the warm-start capability of the
    qpOASES quadratic programming solver. It uses two different
    qp solvers. One to solve for maximized controllable sets and one to
    solve for minimized controllable sets. The wrapper selects which solver
    to use by looking at the optimization direction.

    This solver wrapper also scale data before invoking `qpOASES`.

    If the logger "toppra" is set to debug level, qpoases solvers are
    initialized with PrintLevel.HIGH. Otherwise, these are initialized
    with PrintLevel.NONE

    Currently only support Canonical Linear Constraints.

    Parameters
    ----------
    constraint_list: :class:`.Constraint` []
        The constraints the robot is subjected to.
    path: :class:`.Interpolator`
        The geometric path.
    path_discretization: array
        The discretized path positions.
    disable_check: bool, optional
        Disable check for solution validity. Improve speed by about
        20% but entails the possibility that failure is not reported
        correctly.
    scaling_solverwrapper: bool, optional
        If is True, try to scale the data of each optimization before
        running.  Important: Currently scaling is always done
        regardless of the value of this variable. To be fixed.

    """

    def __init__(
        self,
        constraint_list,
        path,
        path_discretization,
        disable_check=False,
        scaling_solverwrapper=True,
    ):
        if not qpoases_FOUND:
            SolverNotFound("toppra is unable to find any installation of qpoases!")
        super(hotqpOASESSolverWrapper, self).__init__(
            constraint_list, path, path_discretization
        )
        self._disable_check = disable_check

        # First constraint is x + 2 D u <= xnext_max, second is xnext_min <= x + 2D u
        self.nC = 2  # number of Constraints.
        for i, constraint in enumerate(constraint_list):
            if constraint.get_constraint_type() != ConstraintType.CanonicalLinear:
                raise NotImplementedError
            a, b, c, F, v, _, _ = self.params[i]
            if a is not None:
                if constraint.identical:
                    self.nC += F.shape[0]
                else:
                    self.nC += F.shape[1]

        # qpOASES coefficient arrays
        # l <= var <= h
        # lA <= A var <= hA
        self._A = np.zeros((self.nC, self.nV))
        self._lA = -np.ones(self.nC) * QPOASES_INFTY
        self._hA = np.ones(self.nC) * QPOASES_INFTY
        self._l = -np.ones(2) * QPOASES_INFTY
        self._h = np.ones(2) * QPOASES_INFTY

    def setup_solver(self):
        """Initiate two internal solvers for warm-start.
        """
        option = Options()
        if logger.getEffectiveLevel() == logging.DEBUG:
            # option.printLevel = PrintLevel.HIGH
            option.printLevel = PrintLevel.NONE
        else:
            option.printLevel = PrintLevel.NONE
        self.solver_minimizing = SQProblem(self.nV, self.nC)
        self.solver_minimizing.setOptions(option)
        self.solver_maximizing = SQProblem(self.nV, self.nC)
        self.solver_maximizing.setOptions(option)

        self.solver_minimizing_recent_index = -2
        self.solver_maximizing_recent_index = -2

    def close_solver(self):
        self.solver_minimizing = None
        self.solver_maximizing = None

    def solve_stagewise_optim(self, i, H, g, x_min, x_max, x_next_min, x_next_max):
        assert i <= self.N and 0 <= i

        # solve the scaled optimization problem
        #  min    0.5 y^T scale H scale y + g^T scale y
        #  s.t    lA <= A scale y <= hA
        #         l  <=  scale y <= h

        self._l[:] = -QPOASES_INFTY
        self._h[:] = QPOASES_INFTY

        if x_min is not None:
            self._l[1] = max(self._l[1], x_min)
        if x_max is not None:
            self._h[1] = min(self._h[1], x_max)

        if i < self.N:
            delta = self.get_deltas()[i]
            if x_next_min is not None:
                self._A[0] = [-2 * delta, -1]
                self._hA[0] = -x_next_min
            else:
                self._A[0] = [0, 0]
                self._hA[0] = QPOASES_INFTY
            self._lA[0] = -QPOASES_INFTY
            if x_next_max is not None:
                self._A[1] = [2 * delta, 1]
                self._hA[1] = x_next_max
            else:
                self._A[1] = [0, 0]
                self._hA[1] = QPOASES_INFTY
            self._lA[1] = -QPOASES_INFTY
        cur_index = 2
        for j in range(len(self.constraints)):
            a, b, c, F, v, ubound, xbound = self.params[j]

            if a is not None:
                if self.constraints[j].identical:
                    nC_ = F.shape[0]
                    self._A[cur_index : cur_index + nC_, 0] = F.dot(a[i])
                    self._A[cur_index : cur_index + nC_, 1] = F.dot(b[i])
                    self._hA[cur_index : cur_index + nC_] = v - F.dot(c[i])
                    self._lA[cur_index : cur_index + nC_] = -QPOASES_INFTY
                else:
                    nC_ = F[i].shape[0]
                    self._A[cur_index : cur_index + nC_, 0] = F[i].dot(a[i])
                    self._A[cur_index : cur_index + nC_, 1] = F[i].dot(b[i])
                    self._hA[cur_index : cur_index + nC_] = v[i] - F[i].dot(c[i])
                    self._lA[cur_index : cur_index + nC_] = -QPOASES_INFTY
                cur_index = cur_index + nC_
            if ubound is not None:
                self._l[0] = max(self._l[0], ubound[i, 0])
                self._h[0] = min(self._h[0], ubound[i, 1])

            if xbound is not None:
                self._l[1] = max(self._l[1], xbound[i, 0])
                self._h[1] = min(self._h[1], xbound[i, 1])

        # if x_min == x_max, do not solve the 2D linear program, instead, do a line search
        if abs(x_min - x_max) < TINY and H is None and self.get_no_vars() == 2:
            logger.debug("x_min ({:f}) equals x_max ({:f})".format(x_min, x_max))
            u_min = -QPOASES_INFTY
            u_max = QPOASES_INFTY
            for i in range(self._A.shape[0]):
                if self._A[i, 0] > 0:
                    u_max = min(
                        u_max, (self._hA[i] - self._A[i, 1] * x_min) / self._A[i, 0]
                    )
                elif self._A[i, 0] < 0:
                    u_min = max(
                        u_min, (self._hA[i] - self._A[i, 1] * x_min) / self._A[i, 0]
                    )
            if (u_min - u_max) / abs(u_max) > SMALL:  # problem infeasible
                logger.debug(
                    "u_min > u_max by {:f}. Might not be critical. "
                    "Returning failure.".format(u_min - u_max)
                )
                return np.array([np.nan, np.nan])

            if g[0] < 0:
                return np.array([u_max, x_min + 2 * u_max * delta])
            else:
                return np.array([u_min, x_min + 2 * u_min * delta])

        if H is None:
            H = (
                np.ones((self.get_no_vars(), self.get_no_vars())) * 1e-18
            )  # regularization,  very important

        ratio_col1 = 1 / (
            np.sum(np.abs(self._A[2:, 0])) + 1e-5
        )  # the maximum possible value for both ratios is 100000
        ratio_col2 = 1 / (np.sum(np.abs(self._A[2:, 1])) + 1e-5)
        variable_scales = np.array([ratio_col1, ratio_col2])
        # variable_scales = np.array([5000.0, 2000.0])
        variable_scales_mat = np.diag(variable_scales)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                "min ratio col 1 {:f}, col 2 {:f}".format(ratio_col1, ratio_col2)
            )

        # ratio scaling
        self._A = self._A.dot(variable_scales_mat)
        self._l = self._l / variable_scales
        self._h = self._h / variable_scales
        self._g = g * variable_scales
        self._H = variable_scales_mat.dot(H).dot(variable_scales_mat)

        # rows scaling
        row_magnitude = np.sum(np.abs(self._A), axis=1)
        row_scaling_mat = np.diag((row_magnitude + 1) ** (-1))
        self._A = np.dot(row_scaling_mat, self._A)
        self._lA = np.dot(row_scaling_mat, self._lA)
        self._hA = np.dot(row_scaling_mat, self._hA)

        return_value, var = self._solve_optimization(i)

        if return_value == ReturnValue.SUCCESSFUL_RETURN:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("optimal value: {:}".format(var))

            if self._disable_check:
                return var * variable_scales

            # Check for constraint feasibility
            success = (
                np.all(self._l <= var + TINY)
                and np.all(var <= self._h + TINY)
                and np.all(np.dot(self._A, var) <= self._hA + TINY)
                and np.all(np.dot(self._A, var) >= self._lA - TINY)
            )
            if not success:
                # import ipdb; ipdb.set_trace()
                logger.fatal(
                    "Hotstart fails but qpOASES does not report correctly. \n "
                    "var: {:}, lower_bound: {:}, higher_bound{:}".format(
                        var, self._l, self._h
                    )
                )
                # TODO: Investigate why this happen and fix the
                # relevant code (in qpOASES wrapper)
            else:
                return var * variable_scales
        else:
            logger.debug("Optimization fails. qpOASES error code: %d.", return_value)

            if (
                np.all(0 <= self._hA)
                and np.all(0 >= self._lA)
                and np.all(0 <= self._h)
                and np.all(0 >= self._l)
            ):
                logger.fatal(
                    f"(0, 0) satisfies all constraints => error due to numerical errors, \
                        %s, %s, %s, %s, %s",  self._A, self._lA, self._hA, self._l, self._h
                )
            else:
                logger.debug("(0, 0) does not satisfy all constraints.")

        return_value = np.empty(self.get_no_vars())
        return_value[:] = np.nan
        return return_value

    def _solve_optimization(self, i):
        var = np.zeros(self.nV)
        if self._g[1] > 0:  # Choose solver_minimizing
            if abs(self.solver_minimizing_recent_index - i) > 1:
                logger.debug("solver_minimizing [init]")
                return_value = self.solver_minimizing.init(
                    self._H,
                    self._g,
                    self._A,
                    self._l,
                    self._h,
                    self._lA,
                    self._hA,
                    np.array([1000]),
                )
            else:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("solver_minimizing [hotstart]")
                return_value = self.solver_minimizing.hotstart(
                    self._H,
                    self._g,
                    self._A,
                    self._l,
                    self._h,
                    self._lA,
                    self._hA,
                    np.array([1000]),
                )
            self.solver_minimizing_recent_index = i
            self.solver_minimizing.getPrimalSolution(var)
        else:  # Choose solver_maximizing
            if abs(self.solver_maximizing_recent_index - i) > 1:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("solver_maximizing [init]")
                return_value = self.solver_maximizing.init(
                    self._H,
                    self._g,
                    self._A,
                    self._l,
                    self._h,
                    self._lA,
                    self._hA,
                    np.array([1000]),
                )
            else:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("solver_maximizing [hotstart]")
                return_value = self.solver_maximizing.hotstart(
                    self._H,
                    self._g,
                    self._A,
                    self._l,
                    self._h,
                    self._lA,
                    self._hA,
                    np.array([1000]),
                )
            self.solver_maximizing_recent_index = i
            self.solver_maximizing.getPrimalSolution(var)
        return return_value, var



from .constraint import Constraint
from .constraint import ConstraintType, DiscretizationType
import numpy as np


class ConicConstraint(Constraint):
    """Base class for all canonical conic constraints.

    A canonical conic constraint is one with the following form

    .. math::

        (a[i] + da[i]) u + (b[i] + db[i]) x + (c[i] + dc[i]) \leq 0, \\\\
        [da[i, j], db[i, j], dc[i, j]]^\top = P[i, j] u, \|u\|_2 \leq 1,

    where P[i, j] is a 3x3 matrix. Notice that by setting P[i, j] to
    the zero matrix, this is the same as the original constraint.

    Constraints of this form can be translated to conic-quadratic
    constraints. This transformation can be found in [1]. The
    resulting conic-quadratic constraint is given below

    .. math::
        a[i, j]u + b[i, j]x + c[i, j] + \|P[i, j]^T [u, x, 1]^T \|_2 \leq 0,

    where i is the stage index, and j is the constraint index.

    Refs:
    ----
    [1] Ben-Tal, A., & Nemirovski, A. (2001). Lectures on modern convex
        optimization: analysis, algorithms, and engineering applications
        (Vol. 2). Siam.

    """

    def __init__(self):
        self.constraint_type = ConstraintType.CanonicalConic
        self.discretization_type = DiscretizationType.Collocation
        self.n_extra_vars = 0
        self.dof = -1
        self._format_string = ""

    def compute_constraint_params(self, path, gridpoints):
        raise NotImplementedError


class RobustLinearConstraint(ConicConstraint):
    """The simple canonical conic constraint.

    This constraint can be seen as a robustified version of a
    CanonicalLinear constraint. In particular, the perturbations term,
    [\Delta a[i, j], \Delta b[i, j], \Delta c[i, j]] is assumed to lie
    in a centered ellipsoid:

    .. math::

        [\Delta a[i, j], \Delta b[i, j], \Delta c[i, j]]^\\top = diag(ru, rx, rc) \mathbf e,

    where \|\mathbf e\|_2 \leq 1.

    Parameters
    ----------
    cnst: :class:`~toppra.constraint.LinearConstraint`
        The base constraint to robustify.
    ellipsoid_axes_lengths: (3,)array
        Lengths of the axes of the perturbation ellipsoid. Must all be
        non-negative.
    discretization_scheme: :class:`~.constraint.DiscretizationType`
        Constraint discretization scheme to use.

    """

    def __init__(
        self,
        cnst,
        ellipsoid_axes_lengths,
        discretization_scheme=DiscretizationType.Collocation,
    ):
        super(RobustLinearConstraint, self).__init__()
        self.dof = cnst.get_dof()
        assert cnst.get_constraint_type() == ConstraintType.CanonicalLinear
        self.set_discretization_type(discretization_scheme)
        if np.any(np.r_[ellipsoid_axes_lengths] < 0):
            raise ValueError(
                "Perturbation must be non-negative. Input {:}".format(
                    ellipsoid_axes_lengths
                )
            )
        self.base_constraint = cnst
        self.ellipsoid_axes_lengths = ellipsoid_axes_lengths
        self._format_string += (
            "    Robust constraint generated from a canonical linear constraint\n"
        )

    def compute_constraint_params(self, path, gridpoints):
        self.base_constraint.set_discretization_type(self.discretization_type)
        a_, b_, c_, F_, g_, u_, x_ = self.base_constraint.compute_constraint_params(
            path, gridpoints
        )
        N = len(gridpoints) - 1
        if self.base_constraint.identical:
            d = F_.shape[0]  # number of rows
        else:
            d = F_.shape[1]

        a = np.zeros((N + 1, d))
        b = np.zeros((N + 1, d))
        c = np.zeros((N + 1, d))

        if self.base_constraint.identical:
            for i in range(len(gridpoints)):
                a[i, :d] = F_.dot(a_[i])
                b[i, :d] = F_.dot(b_[i])
                c[i, :d] = F_.dot(c_[i]) - g_
        else:
            for i in range(len(gridpoints)):
                a[i, :d] = F_[i].dot(a_[i])
                b[i, :d] = F_[i].dot(b_[i])
                c[i, :d] = F_[i].dot(c_[i]) - g_[i]

        P = np.zeros((N + 1, d + 2, 3, 3))
        diag_ = np.diag(self.ellipsoid_axes_lengths)
        P[:] = diag_
        return a, b, c, P, u_, x_



"""
Path Interpolator
-----------------------

This module implements clases to represent geometric paths and
trajectories.

SplineInterplator
^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.SplineInterpolator
   :members: __call__, dof, path_interval, waypoints

RaveTrajectoryWrapper
^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.RaveTrajectoryWrapper
   :members: __call__, dof, path_interval, waypoints

simplepath.SimplePath
^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.simplepath.SimplePath
   :members: __call__, dof, path_interval

[abstract]AbstractGeometricPath
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: toppra.interpolator.AbstractGeometricPath
   :members: __call__, dof, path_interval, waypoints

[internal]Methods
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autofunction:: toppra.interpolator.propose_gridpoints


"""
from typing import List, Union
import typing as T
import logging
import numpy as np
from scipy.interpolate import UnivariateSpline, CubicSpline, PPoly
from toppra.utils import deprecated
from toppra.constants import FOUND_OPENRAVE

logger = logging.getLogger(__name__)

if FOUND_OPENRAVE:
    import openravepy as orpy  # pylint: disable=import-error


def propose_gridpoints(
        path, max_err_threshold=1e-4, max_iteration=100, max_seg_length=0.05,
        min_nb_points=100
):
    r"""Generate gridpoints that sufficiently cover the given path.

    This function operates in multiple passes through the geometric
    path from the start to the end point. In each pass, for each
    segment, the maximum interpolation error is estimated using the
    following equation:

    .. math::

        err_{est} = 0.5 * \mathrm{max}(\mathrm{abs}(p'' * d_{segment} ^ 2))

    Here :math:`p''` is the second derivative of the path and
    d_segment is the length of the segment. If the estimated error
    :math:`err_{test}` is greater than the given threshold
    `max_err_threshold` then the segment is divided in two half.

    Intuitively, at positions with higher curvature, there must be
    more points in order to improve approximation
    quality. Theoretically toppra performs the best when the proposed
    gridpoint is optimally distributed.

    Arguments
    ---------
    path: :class:`AbstractGeometricPath`
      Input geometric path.
    max_err_threshold: float
      Maximum worstcase error thrshold allowable.
    max_iteration: int
      Maximum number of iterations.
    max_seg_length: float
      All segments length should be smaller than this value.

    Returns
    ----------
    gridpoints_ept: np.ndarray(N,)
      The proposed gridpoints.

    """
    gridpoints_ept = [path.path_interval[0], path.path_interval[1]]
    for iteration in range(max_iteration):
        add_new_points = False
        for idx in range(len(gridpoints_ept) - 1):
            gp_mid = 0.5 * (gridpoints_ept[idx] + gridpoints_ept[idx + 1])
            if (gridpoints_ept[idx + 1] - gridpoints_ept[idx]) > max_seg_length:
                gridpoints_ept.append(gp_mid)
                add_new_points = True
                continue

            dist = gridpoints_ept[idx + 1] - gridpoints_ept[idx]
            max_err = np.max(np.abs(0.5 * path(gp_mid, 2) * dist ** 2))
            if max_err > max_err_threshold:
                add_new_points = True
                gridpoints_ept.append(gp_mid)
                continue

        gridpoints_ept = sorted(gridpoints_ept)
        if not add_new_points:
            break

    while len(gridpoints_ept) < min_nb_points:
        new_pts = []
        for idx in range(len(gridpoints_ept) - 1):
            gp_mid = 0.5 * (gridpoints_ept[idx] + gridpoints_ept[idx + 1])
            new_pts.append(gp_mid)
        gridpoints_ept.extend(new_pts)
        gridpoints_ept = sorted(gridpoints_ept)

    if iteration == max_iteration - 1:
        raise ValueError("Unable to find a good gridpoint for this path.")
    return gridpoints_ept


class AbstractGeometricPath(object):
    """Abstract base class that represents geometric paths.

    Derive geometric paths classes should implement the below abstract
    methods. These methods are expected in different steps of the
    algorithm.

    """

    def __call__(self, path_positions: Union[float, np.ndarray], order: int = 0) -> np.ndarray:
        """Evaluate the path at given positions.

        Parameters
        ----------
        path_positions: float or np.ndarray
            Path positions to evaluate the interpolator.
        order: int
            Order of the evaluation call.

            - 0: position
            - 1: first-order derivative
            - 2: second-order derivative

        Returns
        -------
        :
            The evaluated joint positions, velocity or
            accelerations. The shape of the result depends on the
            shape of the input, it is either (N, m) where N is the
            number of path positions and m is the number of
            degree-of-freedom, or (m,).

        """
        raise NotImplementedError

    @property
    def dof(self) -> int:
        """Return the degrees-of-freedom of the path."""
        raise NotImplementedError

    @property
    def path_interval(self):
        """Return the starting and ending path positions.

        Returns
        -------
        np.ndarray(2,)
            The starting and ending path positions.

        """
        raise NotImplementedError

    @property
    def waypoints(self):
        """Tuple[ndarray, ndarray] or None: The path's waypoints if applicable. None otherwise."""
        return None

    def eval(self, ss_sam: Union[float, np.ndarray]):
        """Evaluate the path values."""
        return self(ss_sam, 0)

    def evald(self, ss_sam: Union[float, np.ndarray]):
        """Evaluate the path first-derivatives."""
        return self(ss_sam, 1)

    def evaldd(self, ss_sam: Union[float, np.ndarray]):
        """Evaluate the path second-derivatives."""
        return self(ss_sam, 2)


class RaveTrajectoryWrapper(AbstractGeometricPath):
    """An interpolator that wraps OpenRAVE's :class:`GenericTrajectory`.

    Only trajectories using quadratic interpolation or cubic
    interpolation are supported.  The trajectory is represented as a
    piecewise polynomial. The polynomial could be quadratic or cubic
    depending the interpolation method used by the input trajectory
    object.

    """

    @staticmethod
    def _extract_interpolation_method(spec):
        _interpolation = spec.GetGroupFromName("joint").interpolation
        if _interpolation not in ["quadratic", "cubic"]:
            raise ValueError(
                "This class only handles trajectories with quadratic or cubic interpolation"
            )
        return _interpolation

    def __init__(self, traj, robot):
        # type: (orpy.RaveTrajectory, orpy.Robot) -> None
        """Initialize the Trajectory Wrapper.

        Parameters
        ----------
        traj:
            An OpenRAVE joint trajectory.
        robot:
            An OpenRAVE robot.

        """
        super(RaveTrajectoryWrapper, self).__init__()
        self.traj = traj
        spec = traj.GetConfigurationSpecification()
        self._dof = robot.GetActiveDOF()
        self._interpolation = self._extract_interpolation_method(spec)
        self._duration = traj.GetDuration()

        all_waypoints = traj.GetWaypoints(0, traj.GetNumWaypoints()).reshape(
            traj.GetNumWaypoints(), -1
        )
        valid_wp_indices = [0]
        ss_waypoints: List[float] = [0.0]
        for i in range(1, traj.GetNumWaypoints()):
            dt = spec.ExtractDeltaTime(all_waypoints[i])
            if dt > 1e-5:  # If delta is too small, skip it.
                valid_wp_indices.append(i)
                ss_waypoints.append(ss_waypoints[-1] + dt)
        self.ss_waypoints = np.array(ss_waypoints)
        n_waypoints = len(valid_wp_indices)

        def _extract_waypoints(order):
            return np.array(
                [
                    spec.ExtractJointValues(
                        all_waypoints[i], robot, robot.GetActiveDOFIndices(), order
                    )
                    for i in valid_wp_indices
                ]
            )

        def _make_ppoly():
            if n_waypoints == 1:
                waypoints = _extract_waypoints(0)
                pp_coeffs = np.zeros((1, 1, self.dof))
                for idof in range(self.dof):
                    pp_coeffs[0, 0, idof] = waypoints[0, idof]
                return PPoly(pp_coeffs, [0, 1])

            if self._interpolation == "quadratic":
                waypoints = _extract_waypoints(0)
                waypoints_d = _extract_waypoints(1)
                waypoints_dd = []
                for i in range(n_waypoints - 1):
                    qdd = (waypoints_d[i + 1] - waypoints_d[i]) / (
                        self.ss_waypoints[i + 1] - self.ss_waypoints[i]
                    )
                    waypoints_dd.append(qdd)
                waypoints_dd = np.array(waypoints_dd)

                # Fill the coefficient matrix for scipy.PPoly class
                pp_coeffs = np.zeros((3, n_waypoints - 1, self.dof))
                for idof in range(self.dof):
                    for iseg in range(n_waypoints - 1):
                        pp_coeffs[:, iseg, idof] = [
                            waypoints_dd[iseg, idof] / 2,
                            waypoints_d[iseg, idof],
                            waypoints[iseg, idof],
                        ]
                return PPoly(pp_coeffs, self.ss_waypoints)

            if self._interpolation == "cubic":
                waypoints = _extract_waypoints(0)
                waypoints_d = _extract_waypoints(1)
                waypoints_dd = _extract_waypoints(2)
                waypoints_ddd = []
                for i in range(n_waypoints - 1):
                    qddd = (waypoints_dd[i + 1] - waypoints_dd[i]) / (
                        self.ss_waypoints[i + 1] - self.ss_waypoints[i]
                    )
                    waypoints_ddd.append(qddd)
                waypoints_ddd = np.array(waypoints_ddd)

                # Fill the coefficient matrix for scipy.PPoly class
                pp_coeffs = np.zeros((4, n_waypoints - 1, self.dof))
                for idof in range(self.dof):
                    for iseg in range(n_waypoints - 1):
                        pp_coeffs[:, iseg, idof] = [
                            waypoints_ddd[iseg, idof] / 6,
                            waypoints_dd[iseg, idof] / 2,
                            waypoints_d[iseg, idof],
                            waypoints[iseg, idof],
                        ]
                return PPoly(pp_coeffs, self.ss_waypoints)
            raise ValueError("An error has occured. Unable to form PPoly.")

        self.ppoly = _make_ppoly()

    @deprecated
    def get_duration(self):
        """Return the path's duration."""
        return self.duration

    @deprecated
    def get_dof(self):  # type: () -> int
        """Return the path's dof."""
        return self.dof

    @property
    def duration(self):
        """Return the duration of the path."""
        return self._duration

    @property
    def path_interval(self):
        """Return the start and end points."""
        return np.array([0, self._duration])

    @property
    def dof(self):
        return self._dof

    def __call__(self, ss_sam, order=0):
        if order == 0:
            return self.eval(ss_sam)
        if order == 1:
            return self.evald(ss_sam)
        if order == 2:
            return self.evaldd(ss_sam)
        raise ValueError("Order must be 0, 1 or 2.")

    def eval(self, ss_sam):
        """Evalute path postition."""
        return self.ppoly(ss_sam)

    def evald(self, ss_sam):
        """Evalute path velocity."""
        return self.ppoly.derivative()(ss_sam)

    def evaldd(self, ss_sam):
        """Evalute path acceleration."""
        return self.ppoly.derivative(2)(ss_sam)


class SplineInterpolator(AbstractGeometricPath):
    """Interpolate the given waypoints by cubic spline.

    This interpolator is implemented as a simple wrapper over scipy's
    CubicSpline class.

    Parameters
    ----------
    ss_waypoints: np.ndarray(m,)
        Path positions of the waypoints.
    waypoints: np.ndarray(m, d)
        Waypoints.
    bc_type: optional
        Boundary conditions of the spline. Can be 'not-a-knot',
        'clamped', 'natural' or 'periodic'.

        - 'not-a-knot': The most default option, return the most naturally
          looking spline.
        - 'clamped': First-order derivatives of the spline at the two
          end are clamped at zero.

        See scipy.CubicSpline documentation for more details.

    """

    def __init__(
        self, 
        ss_waypoints, 
        waypoints, 
        bc_type: str="not-a-knot"
    ) -> None:
        super(SplineInterpolator, self).__init__()
        self.ss_waypoints = np.array(ss_waypoints)  # type: np.ndarray
        self._q_waypoints = np.array(waypoints)  # type: np.ndarray
        assert self.ss_waypoints.shape[0] == self._q_waypoints.shape[0]

        self.cspl: T.Union[T.Callable[[T.Any], T.Any], CubicSpline]
        self.cspld: T.Union[T.Callable[[T.Any], T.Any], CubicSpline]
        if len(ss_waypoints) == 1:

            def _1dof_cspl(s):
                try:
                    ret = np.zeros((len(s), self.dof))
                    ret[:, :] = self._q_waypoints[0]
                except TypeError:
                    ret = self._q_waypoints[0]
                return ret

            def _1dof_cspld(s):
                try:
                    ret = np.zeros((len(s), self.dof))
                except TypeError:
                    ret = np.zeros(self.dof)
                return ret

            self.cspl = _1dof_cspl
            self.cspld = _1dof_cspld
            self.cspldd = _1dof_cspld
        else:
            self.cspl = CubicSpline(ss_waypoints, waypoints, bc_type=bc_type)
            self.cspld = self.cspl.derivative()
            self.cspldd = self.cspld.derivative()

    def __call__(self, path_positions, order=0):
        if order == 0:
            return self.cspl(path_positions)
        if order == 1:
            return self.cspld(path_positions)
        if order == 2:
            return self.cspldd(path_positions)
        raise ValueError(f"Invalid order {order}")

    @property
    def waypoints(self):
        """Tuple[np.ndarray, np.ndarray]: Return the waypoints.

        The first element is the positions, the second element is the
        array of waypoints.

        """
        return self.ss_waypoints, self._q_waypoints

    @deprecated
    def get_duration(self):
        """Return the path's duration."""
        return self.duration

    @property
    def duration(self):
        """Return the duration of the path."""
        return self.ss_waypoints[-1] - self.ss_waypoints[0]

    @property
    def path_interval(self):
        """Return the start and end points."""
        return np.array([self.ss_waypoints[0], self.ss_waypoints[-1]])

    @deprecated
    def get_path_interval(self):
        """Return the path interval."""
        return self.path_interval

    @property
    def dof(self):
        if np.isscalar(self._q_waypoints[0]):
            return 1
        return self._q_waypoints[0].shape[0]

    def compute_rave_trajectory(self, robot):
        """Compute an OpenRAVE trajectory equivalent to this trajectory.

        Parameters
        ----------
        robot:
            Openrave robot.

        Returns
        -------
        trajectory:
            Equivalent openrave trajectory.
        """

        traj = orpy.RaveCreateTrajectory(robot.GetEnv(), "")
        spec = robot.GetActiveConfigurationSpecification("cubic")
        spec.AddDerivativeGroups(1, False)
        spec.AddDerivativeGroups(2, True)

        traj.Init(spec)
        deltas = [0]
        for i in range(len(self.ss_waypoints) - 1):
            deltas.append(self.ss_waypoints[i + 1] - self.ss_waypoints[i])
        if len(self.ss_waypoints) == 1:
            q = self(0)
            qd = self(0, 1)
            qdd = self(0, 2)
            traj.Insert(traj.GetNumWaypoints(), list(q) + list(qd) + list(qdd) + [0])
        else:
            qs = self(self.ss_waypoints)
            qds = self(self.ss_waypoints, 1)
            qdds = self(self.ss_waypoints, 2)
            for (q, qd, qdd, dt) in zip(qs, qds, qdds, deltas):
                traj.Insert(
                    traj.GetNumWaypoints(),
                    q.tolist() + qd.tolist() + qdd.tolist() + [dt],
                )
        return traj


class UnivariateSplineInterpolator(AbstractGeometricPath):
    """ Smooth given wayspoints by a cubic spline.

    This is a simple wrapper over `scipy.UnivariateSplineInterpolator`
    class.

    Parameters
    ----------
    ss_waypoints: ndarray
        Path positions of the waypoints.
    waypoints: ndarray
        The waypoints.
    """

    def __init__(self, ss_waypoints, waypoints):
        super(UnivariateSplineInterpolator, self).__init__()
        assert ss_waypoints[0] == 0, "First index must equals zero."
        self.ss_waypoints = np.array(ss_waypoints)
        self._q_waypoints = np.array(waypoints)
        if np.isscalar(waypoints[0]):
            self._dof = 1
        else:
            self._dof = waypoints[0].shape[0]
        assert self.ss_waypoints.shape[0] == self._q_waypoints.shape[0]
        self.uspl = []
        for i in range(self.dof):
            self.uspl.append(
                UnivariateSpline(self.ss_waypoints, self._q_waypoints[:, i])
            )
        self.uspld = [spl.derivative() for spl in self.uspl]
        self.uspldd = [spl.derivative() for spl in self.uspld]

    @property
    def dof(self):
        return self._dof

    @property
    def path_interval(self):
        """Return the path interval."""
        return [self.ss_waypoints[0], self.ss_waypoints[-1]]

    def __call__(self, ss_sam, order=0):
        data = []
        if order == 0:
            for spl in self.uspl:
                data.append(spl(ss_sam))
        elif order == 1:
            for spl in self.uspld:
                data.append(spl(ss_sam))
        elif order == 2:
            for spl in self.uspldd:
                data.append(spl(ss_sam))
        return np.array(data).T

    def eval(self, ss_sam):
        """Return the path position."""
        data = []
        for spl in self.uspl:
            data.append(spl(ss_sam))
        return np.array(data).T

    def evald(self, ss_sam):
        """Return the path velocity."""
        data = []
        for spl in self.uspld:
            data.append(spl(ss_sam))
        return np.array(data).T

    def evaldd(self, ss_sam):
        """Return the path acceleration."""
        data = []
        for spl in self.uspldd:
            data.append(spl(ss_sam))
        return np.array(data).T


class PolynomialPath(AbstractGeometricPath):
    """ A class representing polynominal paths.

    If coeff is a 1d array, the polynomial's equation is given by

    .. math::

    coeff[0] + coeff[1] s + coeff[2] s^2 + ...

    If coeff is a 2d array, the i-th joint position is the polynomial

    .. math::

    coeff[i, 0] + coeff[i, 1] s + coeff[i, 2] s^2 + ...
    """

    def __init__(self, coeff, s_start=0.0, s_end=1.0) -> None:
        """Initialize the polynomial path.

        Parameters
        ----------
        coeff
            Coefficients of the polynomials.
        s_start
            Starting path position.
        s_end
            Ending path position.
        """
        super(PolynomialPath, self).__init__()

        self.coeff = coeff
        self.s_end = s_end
        self.s_start = s_start
        if np.isscalar(self.coeff[0]):
            self._dof = 1
            self.poly = [np.polynomial.Polynomial(self.coeff)]
            self.coeff = np.array(self.coeff).reshape(1, -1)
        else:
            self._dof = len(self.coeff)
            self.poly = [
                np.polynomial.Polynomial(self.coeff[i]) for i in range(self._dof)
            ]

        self.polyd = [poly.deriv() for poly in self.poly]
        self.polydd = [poly.deriv() for poly in self.polyd]

    def __call__(self, path_positions, order=0):
        if order == 0:
            return self.eval(path_positions)
        if order == 1:
            return self.evald(path_positions)
        if order == 2:
            return self.evaldd(path_positions)
        raise ValueError(f"Invalid order {order}")

    @property
    def dof(self):
        return self._dof

    @property
    def duration(self):
        """Return the duration of the path."""
        return self.s_end - self.s_start

    @property
    def path_interval(self):
        return np.array([self.s_start, self.s_end])

    @deprecated
    def get_path_interval(self):
        """Return the path interval."""
        return self.path_interval

    @deprecated
    def get_duration(self):
        """Return the path duration."""
        return self.duration

    @deprecated
    def get_dof(self):
        """Return the path's dof."""
        return self.dof

    def eval(self, ss_sam):
        """Return the path position."""
        res = [poly(np.array(ss_sam)) for poly in self.poly]
        if self.dof == 1:
            return np.array(res).flatten()
        return np.array(res).T

    def evald(self, ss_sam):
        """Return the path velocity."""
        res = [poly(np.array(ss_sam)) for poly in self.polyd]
        if self.dof == 1:
            return np.array(res).flatten()
        return np.array(res).T

    def evaldd(self, ss_sam):
        """Return the path acceleration."""
        res = [poly(np.array(ss_sam)) for poly in self.polydd]
        if self.dof == 1:
            return np.array(res).flatten()
        return np.array(res).T



"""
toppra.parametrizer
^^^^^^^^^^^^^^^^^^^^^

This module contains classes that produce the output trajectories,
given the input path and the time parametrization.

"""
import logging
import typing as T
import numpy as np
from toppra.interpolator import AbstractGeometricPath, SplineInterpolator
from toppra.exceptions import ToppraError
from toppra.constants import TINY

try:
    import matplotlib.pyplot as plt
except ImportError:
    pass

logger = logging.getLogger(__name__)

class ParametrizeConstAccel(AbstractGeometricPath):
    """Compute output traj under constant acceleration assumption.
    """

    def __init__(
        self,
        path: AbstractGeometricPath,
        gridpoints: np.ndarray,
        velocities: np.ndarray,
    ) -> None:
        self._path = path
        self._ss: np.ndarray = np.array(gridpoints)
        self._velocities: np.ndarray = np.array(velocities)
        self._xs: np.ndarray = self._velocities ** 2
        self._ts: T.Optional[np.ndarray] = None
        self._us: T.Optional[np.ndarray] = None

        # preconditions
        assert self._ss.shape[0] == self._velocities.shape[0]
        assert len(self._ss.shape) == 1
        assert np.all(self._velocities >= 0)

        self._process_parametrization()

    @property
    def dof(self) -> int:
        """Return the degrees-of-freedom of the path."""
        return self._path.dof

    def _process_parametrization(self):
        ts = [0]
        us = []
        for i in range(self._ss.shape[0] - 1):
            us.append(
                0.5 * (self._xs[i + 1] - self._xs[i]) / (self._ss[i + 1] - self._ss[i])
            )
            ts.append(
                ts[-1]
                + 2
                * (self._ss[i + 1] - self._ss[i])
                / (self._velocities[i] + self._velocities[i + 1])
            )
        self._ts = np.array(ts)
        self._us = np.array(us)

    @property
    def path_interval(self) -> np.ndarray:
        if self._ts is None:
            logger.warning("Unable to find _ts. Processing fails not does not run.")
            raise ValueError("Internal error occur.")
        return np.array([self._ts[0], self._ts[-1]])

    @property
    def duration(self) -> float:
        """Return the path duration."""
        return self.path_interval[1] - self.path_interval[0]

    def __call__(self, ts, order=0):
        scalar = False
        if isinstance(ts, (int, float)):
            ts = np.array([ts], dtype=float)
            scalar = True
        ss, vs, us = self._eval_params(ts)
        if order == 0:
            out = self._path(ss)
        elif order == 1:
            out = np.multiply(self._path(ss, 1), vs[:, np.newaxis])
        elif order == 2:
            out = (np.multiply(self._path(ss, 2), vs[:, np.newaxis] ** 2) +
                   np.multiply(self._path(ss, 1), us[:, np.newaxis]))
        else:
            raise ToppraError(f"Order {order} is not supported.")
        if scalar:
            return out[0]
        return out

    def _eval_params(self, ts: np.ndarray) -> T.Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Return the array of path positions, velocities and accels.

        Parameters
        -----------
        ts:
            Array of time instances.

        Returns
        ---------
        :
            Positions, velocity and accelerations.
        """
        assert self._us is not None
        assert self._ts is not None
        indices = np.searchsorted(self._ts, ts, side="right") - 1
        ss = []
        vs = []
        us = []
        for idx, t in zip(indices, ts):
            if idx == len(self._us):
                idx -= 1
            dt = t - self._ts[idx]
            us.append(self._us[idx])
            vs.append(self._velocities[idx] + dt * self._us[idx])
            ss.append(
                self._ss[idx]
                + dt * self._velocities[idx]
                + 0.5 * dt ** 2 * self._us[idx]
            )
        return np.array(ss), np.array(vs), np.array(us)

    def plot_parametrization(self, show: bool = False, n_sample: int = 500) -> None:
        """Plot the output parametrization and show it."""

        # small decrement to make sure all indices are valid
        ts = np.linspace(self.path_interval[0], self.path_interval[1], n_sample)
        ss, vs, _ = self._eval_params(ts)
        qs = self(ts, 0)

        plt.subplot(2, 2, 1)
        plt.plot(ts, ss, label="s(t)")
        plt.plot(self._ts, self._ss, "o", label="input")
        plt.title("path(time)")
        plt.legend()
        plt.subplot(2, 2, 2)
        plt.plot(ss, vs, label="v(s)")
        plt.plot(self._ss, self._velocities, "o", label="input")
        plt.title("velocity(path)")
        plt.legend()
        plt.subplot(2, 2, 3)
        plt.plot(ts, qs)
        plt.title("retimed path")
        plt.subplot(2, 2, 4)
        ss_dense = np.linspace(self._ss[0], self._ss[-1], n_sample)
        plt.plot(ss_dense, self._path(ss_dense))
        plt.title("original path")
        plt.tight_layout()
        if show:
            plt.show()


class ParametrizeSpline(SplineInterpolator):
    """Return output trajetory via sphine interpolation.

    This class computes the time and position at each gridpoint, then
    fit and return a CubicSpline (continuous first and second
    derivatives). Note that the boundary conditions are: first
    derivatives at the start and end of the path equal q(s)' * s'.

    """

    def __init__(self, path, gridpoints, velocities):
        # Gridpoint time instances
        t_grid = np.zeros_like(gridpoints)
        skip_ent = []
        for i in range(1, len(t_grid)):
            sd_average = (velocities[i - 1] + velocities[i]) / 2
            delta_s = gridpoints[i] - gridpoints[i - 1]
            if sd_average > TINY:
                delta_t = delta_s / sd_average
            else:
                delta_t = 5  # If average speed is too slow.
            t_grid[i] = t_grid[i - 1] + delta_t
            if delta_t < TINY:  # if a time increment is too small, skip.
                skip_ent.append(i)
        t_grid = np.delete(t_grid, skip_ent)
        gridpoints = np.delete(gridpoints, skip_ent)
        q_grid = path(gridpoints)

        super(ParametrizeSpline, self).__init__(
            t_grid,
            q_grid,
            (
                (1, path(path.path_interval[0], 1) * velocities[0]),
                (1, path(path.path_interval[1], 1) * velocities[-1]),
            ),
        )



"""Joint acceleration constraint."""
import numpy as np
from .linear_constraint import LinearConstraint, canlinear_colloc_to_interpolate
from ..constraint import DiscretizationType
from ..interpolator import AbstractGeometricPath


class JointAccelerationConstraint(LinearConstraint):
    """The Joint Acceleration Constraint class.

    A joint acceleration constraint is given by

    .. math ::

                \ddot{\mathbf{q}}_{min} & \leq \ddot{\mathbf q}
                                                    &\leq \ddot{\mathbf{q}}_{max} \\\\
                \ddot{\mathbf{q}}_{min} & \leq \mathbf{q}'(s_i) u_i + \mathbf{q}''(s_i) x_i
                                                    &\leq \ddot{\mathbf{q}}_{max}

    where :math:`u_i, x_i` are respectively the path acceleration and
    path velocity square at :math:`s_i`. For more detail see :ref:`derivationKinematics`.

    Rearranging the above pair of vector inequalities into the form
    required by :class:`LinearConstraint`, we have:

    - :code:`a[i]` := :math:`\mathbf q'(s_i)`
    - :code:`b[i]` := :math:`\mathbf q''(s_i)`
    - :code:`F` := :math:`[\mathbf{I}, -\mathbf I]^T`
    - :code:`h` := :math:`[\ddot{\mathbf{q}}_{max}^T, -\ddot{\mathbf{q}}_{min}^T]^T`
    """

    def __init__(self, alim, discretization_scheme=DiscretizationType.Interpolation):
        """Initialize the joint acceleration class.

        Parameters
        ----------
        alim: array
            Shape (dof, 2). The lower and upper acceleration bounds of the
            j-th joint are alim[j, 0] and alim[j, 1] respectively.

        discretization_scheme: :class:`.DiscretizationType`
            Can be either Collocation (0) or Interpolation
            (1). Interpolation gives more accurate results with slightly
            higher computational cost.
        """
        super(JointAccelerationConstraint, self).__init__()
        alim = np.array(alim, dtype=float)
        if np.isnan(alim).any():
            raise ValueError("Bad velocity given: %s" % alim)
        if len(alim.shape) == 1:
            self.alim = np.vstack((-np.array(alim), np.array(alim))).T
        else:
            self.alim = np.array(alim, dtype=float)
        self.dof = self.alim.shape[0]
        self.set_discretization_type(discretization_scheme)

        assert self.alim.shape[1] == 2, "Wrong input shape."
        self._format_string = "    Acceleration limit: \n"
        for i in range(self.alim.shape[0]):
            self._format_string += "      J{:d}: {:}".format(i + 1, self.alim[i]) + "\n"
        self.identical = True

    def compute_constraint_params(
        self, path: AbstractGeometricPath, gridpoints: np.ndarray, *args, **kwargs
    ):
        if path.dof != self.dof:
            raise ValueError(
                "Wrong dimension: constraint dof ({:d}) not equal to path dof ({:d})".format(
                    self.dof, path.dof
                )
            )
        ps_vec = (path(gridpoints, order=1)).reshape((-1, path.dof))
        pss_vec = (path(gridpoints, order=2)).reshape((-1, path.dof))
        dof = path.dof
        F_single = np.zeros((dof * 2, dof))
        g_single = np.zeros(dof * 2)
        g_single[0:dof] = self.alim[:, 1]
        g_single[dof:] = -self.alim[:, 0]
        F_single[0:dof, :] = np.eye(dof)
        F_single[dof:, :] = -np.eye(dof)
        if self.discretization_type == DiscretizationType.Collocation:
            return (
                ps_vec,
                pss_vec,
                np.zeros_like(ps_vec),
                F_single,
                g_single,
                None,
                None,
            )
        elif self.discretization_type == DiscretizationType.Interpolation:
            return canlinear_colloc_to_interpolate(
                ps_vec,
                pss_vec,
                np.zeros_like(ps_vec),
                F_single,
                g_single,
                None,
                None,
                gridpoints,
                identical=True,
            )
        else:
            raise NotImplementedError("Other form of discretization not supported!")



# cython: embedsignature=True
import numpy as np
cimport numpy as np
from libc.math cimport abs, pow, isnan
cimport cython
from cpython.array cimport array, clone

ctypedef np.int_t INT_t
ctypedef np.float_t FLOAT_t

from ..constraint import ConstraintType

cdef inline double dbl_max(double a, double b): return a if a > b else b
cdef inline double dbl_min(double a, double b): return a if a < b else b

# constants
cdef double TINY = 1e-10
cdef double SMALL = 1e-8

# bounds on variable used in seidel solver wrapper. u and x at every
# stage is constrained to stay within this range.
cdef double VAR_MIN = -100000000  
cdef double VAR_MAX =  100000000

# absolute largest value that a variable can have. This bound should
# be never be reached, however, in order for the code to work properly.
cdef double INF = 10000000000  

cdef double NAN = float("NaN")



cdef struct LpSol:
    # A struct that contains a solution to a Linear Program computed
    # using codes in this source file.
    int result
    double optval
    double[2] optvar
    int[2] active_c


def solve_lp1d(double[:] v, a, b, double low, double high):
    """Solve a Linear Program with 1 variable.

    See func:`cy_solve_lp1d` for more details.

    NOTE: Here a, b, are not typed because a valid input is the empty
    list. However, empty lists are not valid inputs to the cython
    version of this function. So, take extra precaution. 

    Another related note is by not statically typing `a` and `b`, this
    function is much slower than what it can be. My current benchmark
    for 1000 inequalities is 9ms, two-times slower than when there is
    static typing.

    """
    if a is None:
        data = cy_solve_lp1d(v, 0, None, None, low, high)
    elif len(a) == 0:
        data = cy_solve_lp1d(v, 0, None, None, low, high)
    else:
        data = cy_solve_lp1d(v, len(a), a, b, low, high)
    return data.result, data.optval, data.optvar[0], data.active_c[0]

def solve_lp2d(double[:] v, double[:] a, double[:] b, double[:] c, double[:] low, double[:] high, INT_t[:] active_c):
    """ Solve a Linear Program with 2 variable.

    See func:`cy_solve_lp2d` for more details.
    """
    cdef:
        INT_t [:] idx_map
        double[:] a_1d
        double[:] b_1d
    if a is not None and len(a) != 0:
        nrows = a.shape[0]
        idx_map = np.zeros_like(a, dtype=int)
        a_1d = np.zeros(nrows + 4)
        b_1d = np.zeros(nrows + 4)
        use_cache = True
    else:
        idx_map = None
        a_1d = None
        b_1d = None
        use_cache = False

    data = cy_solve_lp2d(v, a, b, c, low, high, active_c, use_cache, idx_map, a_1d, b_1d)
    return data.result, data.optval, data.optvar, data.active_c


# @cython.profile(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef LpSol cy_solve_lp1d(double[:] v, int nrows, double[:] a, double[:] b, double low, double high) except *:
    """ Solve the following program
    
            dbl_max   v[0] x + v[1]
            s.t.  a x + b <= 0
                  low <= x <= high

    """
    cdef:
        double cur_min = low
        int active_c_min = -1
        double cur_max = high
        int active_c_max = -2
        unsigned int i
        double cur_x, optvar, optval
        LpSol solution
        double low_1d = VAR_MIN
        double high_1d = VAR_MAX

    for i in range(nrows):
        if a[i] > TINY:
            cur_x = - b[i] / a[i]
            if cur_x < cur_max:
                cur_max = cur_x
                active_c_max = i
        elif a[i] < -TINY:
            cur_x = - b[i] / a[i]
            if cur_x > cur_min:
                cur_min = cur_x
                active_c_min = i
        else:
            # a[i] is approximately zero. do nothing.
            pass

    if cur_min > cur_max:
        solution.result = 0
        return solution

    if abs(v[0]) < TINY or v[0] < 0:
        # optimizing direction is perpencicular to the line, or is
        # pointing toward the negative direction.
        solution.result = 1
        solution.optvar[0] = cur_min
        solution.optval = v[0] * cur_min + v[1]
        solution.active_c[0] = active_c_min
    else:
        solution.result = 1
        solution.optvar[0] = cur_max
        solution.optval = v[0] * cur_max + v[1]
        solution.active_c[0] = active_c_max

    return solution

# @cython.profile(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef LpSol cy_solve_lp2d(double[:] v, double[:] a, double[:] b, double[:] c,
                         double[:] low, double[:] high,
                         INT_t[:] active_c, bint use_cache,
                         INT_t [:] index_map, double[:] a_1d, double[:] b_1d) except *:
    """ Solve a LP with two variables.

    The LP is specified as follow:
         dbl_max    v^T [x 1]
            s.t.    a x[0] + b x[1] + c <= 0
                    low <= x <= high

    NOTE: A possible optimization for this function is pruning linear
    constraints that are clearly infeasible. This is not implemented
    because in my current code, the bottleneck is not in solving
    TOPP-RA but in setting up the parameters.

    Parameters
    ----------
    v: double memoryview
    a: double memoryview
    b: double memoryview
    c: double memoryview
    low: double memoryview
    high: double memoryview
    active_c: int memoryview
        Contains (2) indicies of rows in a, b, c that are likely the
        active constraints at the optimal solution.
    use_cache: bool
    index_map: int memoryview
        A view to a pre-allocated integer array, to map from
        [1,...,nrows] to the considered entries. This array is created
        to avoid the cost of initializing a new array.
    a_1d: double memoryview
        A view to an initialized array. This array is created to avoid
        the cost of initializing a new array.
    b_1d: double memoryview
        A view to an initialized array. This array is created to avoid
        the cost of initializing a new array.

    Returns
    -------
    solution: LpSol
        A struct that contains the ouput. In the case of
        infeasibility, the struct has its result field equals 0. Other
        fields are not meaningful.

    """
    cdef:
        # number of working set recomputation
        unsigned int nrows = a.shape[0], nrows_1d, n_wsr = 0  
        unsigned int i, k, j
        double[2] cur_optvar, zero_prj
        double[2] d_tan  # vector parallel to the line
        double[2] v_1d_  # optimizing direction
        double [:] v_1d = v_1d_
        double aj, bj, cj  # temporary symbols

        # absolute bounds used in solving the 1 dimensional
        # optimization sub-problems. These bounds needs to be very
        # large, so that they are never active at the optimization
        # solution of these 1D subproblem..
        double low_1d = - INF  
        double high_1d = INF
        LpSol sol, sol_1d

    v_1d_[:] = [0, 0]
    # print all input to the algorithm
    # print "v={:}\n a={:}\n b={:}\n c={:}\n low={:}\n high={:}\n active_c {:}".format(
    #     *map(repr, map(np.asarray,
    #                    [v, a, b, c, low, high, active_c])))

    if not use_cache:
        index_map = np.arange(nrows)
        v_1d = np.zeros(2)  # optimizing direction
        a_1d = np.zeros(nrows + 4)
        b_1d = np.zeros(nrows + 4)
    else:
        assert index_map.shape[0] == nrows

    # handle fixed bounds (low, high). The following convention is
    # adhered to: fixed bounds are assigned the numbers: -1, -2, -3,
    # -4 according to the following order: low[0], high[0], low[1],
    # high[1].
    for i in range(2):
        if low[i] > high[i]:
            sol.result = 0
            return sol
        if v[i] > TINY:
            cur_optvar[i] = high[i]
            if i == 0:
                sol.active_c[0] = -2
            else:
                sol.active_c[1] = -4
        else:
            cur_optvar[i] = low[i]
            if i == 0:
                sol.active_c[0] = -1
            else:
                sol.active_c[1] = -3
    # print("v: {:}".format(repr(np.array(v))))
    # print("opt_var: {:}".format(repr(np.array(cur_optvar))))

    # If active_c contains valid entries, swap the first two indices
    # in index_map to these values.
    cdef unsigned int cur_row = 2
    if active_c[0] >= 0 and active_c[0] < nrows and active_c[1] >= 0 and active_c[1] < nrows and active_c[0] != active_c[1]:
        # active_c contains valid indices
        index_map[0] = active_c[1]
        index_map[1] = active_c[0]
        for i in range(nrows):
            if i != active_c[0] and i != active_c[1]:
                index_map[cur_row] = i
                cur_row += 1
    else:
        for i in range(nrows):
            index_map[i] = i

    # pre-process the inequalities, remove those that are redundant
    cdef cloned_index_map 
    # print(np.array(index_map))

    # handle other constraints in a, b, c
    for k in range(nrows):
        i = index_map[k]
        # if current optimal variable satisfies the i-th constraint, continue
        if a[i] * cur_optvar[0] + b[i] * cur_optvar[1] + c[i] < TINY:
            continue
        # print("a[i] * cur_optvar[0] + b[i] * cur_optvar[1] + c[i] = {:f}".format(
        #     a[i] * cur_optvar[0] + b[i] * cur_optvar[1] + c[i]))
        # print k
        # otherwise, project all constraints on the line defined by (a[i], b[i], c[i])
        n_wsr += 1
        sol.active_c[0] = i
        # project the origin (0, 0) onto the new constraint
        # let ax + by + c=0 b the new constraint
        # let zero_prj be the projected point, one has
        #     zero_prj =  1 / (a^2 + b^2) [a  -b] [-c]
        #                                 [b   a] [ 0]
        # this can be derived using perpendicularity
        # more specifically
        # zero_prj[0] = -ac / (a^2 + b^2), zero_prj[1] = -bc / (a^2 + b^2)
        zero_prj[0] = - a[i] * c[i] / (pow(a[i], 2) + pow(b[i], 2))
        zero_prj[1] = - b[i] * c[i] / (pow(a[i], 2) + pow(b[i], 2))
        d_tan[0] = -b[i]
        d_tan[1] = a[i]
        v_1d[0] = d_tan[0] * v[0] + d_tan[1] * v[1]
        v_1d[1] = 0
        # project 4 + k constraints onto the parallel line. each
        # constraint occupies a row on vectors a_1d, b_1d.
        nrows_1d = 4 + k
        for j in range(nrows_1d):
            # handle low <= x
            if (j == k):
                aj = -1
                bj = 0
                cj = low[0]
            # handle x <= high
            elif (j == k + 1):
                aj = 1
                bj = 0
                cj = -high[0]
            # handle low <= y 
            elif (j == k + 2):
                aj = 0
                bj = -1
                cj = low[1]
            # handle y <= high
            elif (j == k + 3):
                aj = 0
                bj = 1
                cj = -high[1]
            # handle other constraint
            else:
                aj = a[index_map[j]];
                bj = b[index_map[j]];
                cj = c[index_map[j]];
       
            # projective coefficients to the line
            denom = d_tan[0] * aj + d_tan[1] * bj

            # add respective coefficients to a_1d and b_1d
            if denom > TINY:
                t_limit = - (cj + zero_prj[1] * bj + zero_prj[0] * aj) / denom
                a_1d[j] = 1.0
                b_1d[j] = - t_limit
            elif denom < -TINY:
                t_limit = - (cj + zero_prj[1] * bj + zero_prj[0] * aj) / denom
                a_1d[j] = -1.0
                b_1d[j] = t_limit
            else:
                # the curretly considered constraint is parallel to
                # the base one. Check if they are infeasible, in which
                # case return failure immediately.
                if cj + zero_prj[1] * bj + zero_prj[0] * aj > SMALL:
                    sol.result = 0
                    return sol
                # feasible constraints, specify 0 <= 1
                a_1d[j] = 0
                b_1d[j] = - 1.0

        # solve the projected, 1 dimensional LP
        sol_1d = cy_solve_lp1d(v_1d, nrows_1d, a_1d, b_1d, low_1d, high_1d)
        
        # 1d lp infeasible
        if sol_1d.result == 0:
            sol.result = 0
            return sol
        # feasible, wrapping up
        else:
            # print "v={:}\n a={:}\n b={:}\n low={:}\n high={:}\n nrows={:}".format(
            #     *map(repr, map(np.asarray,
            #                    [v_1d, a_1d, b_1d, low_1d, high_1d, nrows_1d])))
            # compute the current optimal solution
            cur_optvar[0] = zero_prj[0] + sol_1d.optvar[0] * d_tan[0]
            cur_optvar[1] = zero_prj[1] + sol_1d.optvar[0] * d_tan[1]
            # print("opt_var: {:}".format(repr(np.array(cur_optvar))))
            # record the active constraint's index
            if sol_1d.active_c[0] < k:
                sol.active_c[1] = index_map[sol_1d.active_c[0]]
            elif sol_1d.active_c[0] == k:
                sol.active_c[1] = -1
            elif sol_1d.active_c[0] == k + 1:
                sol.active_c[1] = -2
            elif sol_1d.active_c[0] == k + 2:
                sol.active_c[1] = -3
            elif sol_1d.active_c[0] == k + 3:
                sol.active_c[1] = -4
            else:
                # the algorithm should not reach this point. If it
                # does, this means the active constraint at the
                # optimal solution is the fixed bound used in the 1
                # dimensional subproblem. This should not happen
                # though.
                sol.result = 0
                return sol

    # Fill the solution struct
    sol.result = 1
    sol.optvar[0] = cur_optvar[0]
    sol.optvar[1] = cur_optvar[1]
    sol.optval = sol.optvar[0] * v[0] + sol.optvar[1] * v[1] + v[2]
    return sol

cdef class seidelWrapper:
    """ A solver wrapper that implements Seidel's LP algorithm.

    This wrapper can only be used if there is only Canonical Linear
    Constraints.

    Parameters
    ----------
    constraint_list: list of :class:`.Constraint`
        The constraints the robot is subjected to.
    path: :class:`.Interpolator`
        The geometric path.
    path_discretization: array
        The discretized path positions.
    """
    cdef:
        list constraints, _params
        unsigned int N, nV, nC, nCons
        double [:] path_discretization
        double [:] deltas
        double [:] a, b, c, low, high, a_1d, b_1d, v
        object path
        INT_t [:] active_c_up, active_c_down
        INT_t [:] index_map
        double [:, ::1] a_arr, b_arr, c_arr  # mmviews of coefficients of the 2D Lp
        double [:, :] low_arr, high_arr    # mmviews of coefficients of the 2D Lp
        double scaling # path scaling

        INT_t _solve_lp1d  # flag to solve 1d instead of 2d LP
        double [:] bx_c  # mmviews of coefficients of the 1D Lp

        
    # @cython.profile(True)
    def __init__(self, list constraint_list, path, path_discretization, solve_lp1d = 0):
        """ A solver wrapper that implements Seidel's LP algorithm.

        This wrapper can only be used if there is only Canonical Linear
        Constraints.

        Parameters
        ----------
        constraint_list: list of :class:`.Constraint`
            The constraints the robot is subjected to.
        path: :class:`.Interpolator`
            The geometric path.
        path_discretization: array
            The discretized path positions.
        solve_lp1d: int
            if solve_lp1d > 0, solve 1D LP instead of the full 2D LP when possible: when x_min == x_max, solve for u as the variable.
        """
        self.constraints = constraint_list
        self.path = path
        path_discretization = np.array(path_discretization)
        self.path_discretization = path_discretization
        self.N = len(path_discretization) - 1  # Number of stages. Number of point is _N + 1
        self.deltas = path_discretization[1:] - path_discretization[:-1]

        self._solve_lp1d = solve_lp1d
        cdef unsigned int cur_index, j, i, k

        # handle constraint parameters
        nCons = len(constraint_list)
        self.nCons = nCons
        self.nC = 2
        self.nV = 2
        self._params = []
        for i in range(nCons):
            if self.constraints[i].get_constraint_type() != ConstraintType.CanonicalLinear:
                raise NotImplementedError
            a, b, c, F, v, ubnd, xbnd = self.constraints[i].compute_constraint_params(
                self.path, path_discretization)
            if a is not None:
                if self.constraints[i].identical:
                    self.nC += F.shape[0]
                else:
                    self.nC += F.shape[1]
            self._params.append((a, b, c, F, v, ubnd, xbnd))

        # init constraint coefficients for the 2d lps, which are
        # self.a_arr, self.b_arr, self.c_arr, self.low_arr,
        # self.high_arr. The first dimensions of these array all equal
        # N + 1.
        self.a_arr = np.zeros((self.N + 1, self.nC))
        self.b_arr = np.zeros((self.N + 1, self.nC))
        self.c_arr = np.zeros((self.N + 1, self.nC))
        self.low_arr = np.ones((self.N + 1, 2)) * VAR_MIN
        self.high_arr = np.ones((self.N + 1, 2)) * VAR_MAX
        cur_index = 2

        cdef double [:, :] ta, tb, tc
        cdef unsigned int nC_
        for j in range(nCons):
            a_j, b_j, c_j, F_j, h_j, ubound_j, xbound_j = self._params[j]

            if a_j is not None:
                if self.constraints[j].identical:
                    nC_ = F_j.shape[0]
                    # <- Most time consuming code, but this computation seems unavoidable
                    ta = a_j.dot(F_j.T)
                    tb = b_j.dot(F_j.T)
                    tc = c_j.dot(F_j.T) - h_j
                    # <-- End

                    for i in range(self.N + 1):
                        for k in range(nC_):
                            self.a_arr[i, cur_index + k] = ta[i, k]
                            self.b_arr[i, cur_index + k] = tb[i, k]
                            self.c_arr[i, cur_index + k] = tc[i, k]
                else:
                    nC_ = F_j.shape[1]
                    for i in range(self.N + 1):
                        tai = np.dot(F_j[i], a_j[i])
                        tbi = np.dot(F_j[i], b_j[i])
                        tci = np.dot(F_j[i], c_j[i]) - h_j[i]
                        for k in range(nC_):
                            self.a_arr[i, cur_index + k] = tai[k]
                            self.b_arr[i, cur_index + k] = tbi[k]
                            self.c_arr[i, cur_index + k] = tci[k]
                cur_index += nC_

            if ubound_j is not None:
                for i in range(self.N + 1):
                    self.low_arr[i, 0] = dbl_max(self.low_arr[i, 0], ubound_j[i, 0])
                    self.high_arr[i, 0] = dbl_min(self.high_arr[i, 0], ubound_j[i, 1])

            if xbound_j is not None:
                for i in range(self.N + 1):
                    self.low_arr[i, 1] = dbl_max(self.low_arr[i, 1], xbound_j[i, 0])
                    self.high_arr[i, 1] = dbl_min(self.high_arr[i, 1], xbound_j[i, 1])
                    
        # init constraint coefficients for the 1d LPs
        self.a_1d = np.zeros(self.nC + 4)
        self.b_1d = np.zeros(self.nC + 4)
        self.index_map = np.zeros(self.nC, dtype=int)
        self.active_c_up = np.zeros(2, dtype=int)
        self.active_c_down = np.zeros(2, dtype=int)
        self.v = np.zeros(3)

        # cache for b*x + c to solve 1d LP when x_min == x_max
        self.bx_c  = np.zeros(self.nC)

    def get_no_vars(self):
        return self.nV

    def get_no_stages(self):
        return self.N

    def get_deltas(self):
        return np.asarray(self.deltas)

    @property
    def params(self):
        return self._params

    # @cython.profile(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef solve_stagewise_optim(self, unsigned int i, H, np.ndarray g, double x_min, double x_max, double x_next_min, double x_next_max):
        """Solve a stage-wise linear optimization problem.

        The linear optimization problem is described below.

        .. math::
            \\text{min  }  & [u, x] g    \\\\
            \\text{s.t.  } & [u, x] \\text{ is feasible at stage } i \\\\
                           & x_{min} \leq x \leq x_{max}             \\\\
                           & x_{next, min} \leq x + 2 \Delta_i u \leq x_{next, max},

        Parameters
        ----------
        i: int
            The stage index. See notes for details on each variable.
        H: array or None
            This term is not used and is neglected.
        g: (d,)array
            The linear term.
        x_min: float
            If not specified, set to NaN.
        x_max: float
            If not specified, set to NaN.
        x_next_min: float
            If not specified, set to NaN.
        x_next_max: float
            If not specified, set to NaN.

        Returns
        -------
        double C array or list
             If successes, return an array containing the optimal
             variable.  Since NaN is also a valid double, this list
             contains NaN if the optimization problem is infeasible.
        """
        assert i <= self.N and 0 <= i

        # fill coefficient
        cdef:
            unsigned int k, cur_index = 0, j, nC  # indices
            double [2] var  # Result
            LpSol  # Solution struct to hold the 2D or 1D LP result
            double low_arr[2], high_arr[2]
        low_arr[0] = self.low_arr[i, 0]
        low_arr[1] = self.low_arr[i, 1]
        high_arr[0] = self.high_arr[i, 0]
        high_arr[1] = self.high_arr[i, 1]

        # handle x_min <= x_i <= x_max
        if not isnan(x_min):
            low_arr[1] = dbl_max(low_arr[1], x_min)
        if not isnan(x_max):
            high_arr[1] = dbl_min(high_arr[1], x_max)

        # handle x_next_min <= 2 delta u + x_i <= x_next_max
        if i < self.N:
            if isnan(x_next_min):
                self.a_arr[i, 0] = 0
                self.b_arr[i, 0] = 0
                self.c_arr[i, 0] = -1
            else:
                self.a_arr[i, 0] = - 2 * self.deltas[i]
                self.b_arr[i, 0] = - 1.0
                self.c_arr[i, 0] = x_next_min
            if isnan(x_next_max):
                self.a_arr[i, 1] = 0
                self.b_arr[i, 1] = 0
                self.c_arr[i, 1] = -1
            else:
                self.a_arr[i, 1] = 2 * self.deltas[i]
                self.b_arr[i, 1] = 1.0
                self.c_arr[i, 1] = - x_next_max
        else:
            # at the last stage, neglect this constraint
            self.a_arr[i, 0:2] = 0
            self.b_arr[i, 0:2] = 0
            self.c_arr[i, 0:2] = -1

        # objective function
        self.v[0] = - g[0]
        self.v[1] = - g[1]

        if x_min == x_max and self._solve_lp1d > 0:
            # solve 1D Lp
            for ic in range(self.nC):
                self.bx_c[ic] = self.b_arr[i, ic] * x_min + self.c_arr[i, ic]
            self.v[1] = - g[1] * x_min
            solution = cy_solve_lp1d(self.v[0:2], self.nC, self.a_arr[i], self.bx_c, low_arr[0], high_arr[0])

            if solution.result == 0:
                # print("upper solver fails")
                var[0] = NAN
                var[1] = NAN
            else:
                var[0] = solution.optvar[0]
                var[1] = x_min
                # TODO is the following correct?
                if g[1] > 0:  
                    self.active_c_up[0] = solution.active_c[0]
                else:
                    self.active_c_down[0] = solution.active_c[0]
            return var

        # warmstarting feature: one in two solvers, upper and lower,
        # is be selected depending on the sign of g[1]

        # solver selected: upper solver. This is selected when
        # computing the lower bound of the controllable set.
        if g[1] > 0:  
            # print "v={:}\n a={:}\n b={:}\n c={:}\n low={:}\n high={:}\n active_c_up={:}".format(
            #     *map(repr, map(np.asarray,
            #                    [self.v, self.a_arr[i], self.b_arr[i], self.c_arr[i],
            #                     low_arr, high_arr, self.active_c_up])))
            solution = cy_solve_lp2d(self.v, self.a_arr[i], self.b_arr[i], self.c_arr[i],
                                     low_arr, high_arr, self.active_c_up,
                                     True, self.index_map, self.a_1d, self.b_1d)
            if solution.result == 0:
                # print("upper solver fails")
                var[0] = NAN
                var[1] = NAN
            else:
                var[:] = solution.optvar
                self.active_c_up[0] = solution.active_c[0]
                self.active_c_up[1] = solution.active_c[1]

        # solver selected: lower solver. This is when computing the
        # lower bound of the controllable set, or computing the
        # parametrization in the forward pass
        else:
            solution = cy_solve_lp2d(self.v, self.a_arr[i], self.b_arr[i], self.c_arr[i],
                                     low_arr, high_arr, self.active_c_down,
                                     True, self.index_map, self.a_1d, self.b_1d)
            if solution.result == 0:
                # print("lower solver fails")
                # print "v={:}\n a={:}\n b={:}\n c={:}\n low={:}\n high={:}".format(
                    # *map(repr, map(np.asarray,
                                  # [self.v, self.a_arr[i], self.b_arr[i], self.c_arr[i], self.low_arr[i], self.high_arr[i]])))
                var[0] = NAN
                var[1] = NAN
            else:
                var[:] = solution.optvar
                self.active_c_down[0] = solution.active_c[0]
                self.active_c_down[1] = solution.active_c[1]
                # print "v={:}\n a={:}\n b={:}\n c={:}\n low={:}\n high={:}\n result={:}\n-----".format(
                    # *map(repr, map(np.asarray,
                                   # [self.v, self.a_arr[i], self.b_arr[i], self.c_arr[i], self.low_arr[i], self.high_arr[i], var])))
                # print np.asarray(self.active_c_down)
                
        return var

    def setup_solver(self):
        pass

    def close_solver(self):
        pass




from .reachability_algorithm import ReachabilityAlgorithm
from ...constants import LARGE, SMALL
from .. import algorithm as algo
import logging
import numpy as np

logger = logging.getLogger(__name__)


def _compute_duration(xs, deltas):
    """ Compute the duration of the given parametrization.
    """
    sds = np.sqrt(xs)
    t = 0
    for i in range(len(deltas)):
        t += 2 * deltas[i] / (sds[i + 1] + sds[i] + 1e-9)
    return t


class TOPPRAsd(ReachabilityAlgorithm):
    """TOPPRA with specified duration.

    The key technical idea is to compute the **fastest** and the
    **slowest** time parameterizations. Then use bisection search to
    find a convex combination of the parametrizations that has the
    desired duration.

    TODO: The current implementation is inefficient.

    """

    def set_desired_duration(self, desired_duration: float):
        """Set desired duration for the time-parametrization.

        Parameters
        ----------
        desired_duration: 
          The desired duration.
        """
        self.desired_duration = desired_duration

    def compute_parameterization(self, sd_start: float, sd_end: float, return_data: bool=False, atol: float=1e-5):
        """Compute a path parameterization.

        If there is no valid parameterizations, simply return None(s).
        If the desired duration is not achievable, returns the fastest
        or the slowest parameterizations.

        Parameters
        ----------
        sd_start:
            Starting path velocity. Must be positive.
        sd_end:
            Goal path velocity. Must be positive.
        return_data:
            If is True, also return matrix K which contains the controllable sets. Default to False.
        atol:
            Absolute tolerance of duration. Default to 1e-5.

        Returns
        -------
        sdd_vec: array or None
            Shape (N,). Path accelerations.
        sd_vec: array None
            Shape (N+1,). Path velocities.
        v_vec: array or None
            Shape (N,). Auxiliary variables.
        K: array
            Shape (N+1, 2). Return if `return_data` is True. The controllable sets.

        """
        assert sd_end >= 0 and sd_start >= 0, "Path velocities must be positive"
        K = self.compute_controllable_sets(sd_end, sd_end)
        if np.isnan(K).any():
            logger.warn("The set of controllable velocities at the beginning is empty!")
            self._problem_data.return_code = algo.ParameterizationReturnCode.FailUncontrollable
            if return_data:
                return None, None, None, K
            else:
                return None, None, None
        self.problem_data.K = K

        x_start = sd_start ** 2
        if x_start + SMALL < K[0, 0] or K[0, 1] + SMALL < x_start:
            logger.warn("The initial velocity is not controllable. {:f} not in ({:f}, {:f})".format(
                x_start, K[0, 0], K[0, 1]
            ))
            self._problem_data.return_code = algo.ParameterizationReturnCode.FailUncontrollable
            if return_data:
                return None, None, None, K
            else:
                return None, None, None
        N = self.solver_wrapper.get_no_stages()
        deltas = self.solver_wrapper.get_deltas()
        # compute the fastest parametrization
        xs = np.zeros(N + 1)
        xs[0] = x_start
        us = np.zeros(N)
        v_vec = np.zeros((N, self.solver_wrapper.get_no_vars() - 2))

        self.solver_wrapper.setup_solver()
        for i in range(self._N):
            optim_res = self._forward_step(i, xs[i], K[i + 1])
            if np.isnan(optim_res[0]):
                logger.fatal("A numerical error occurs: The instance is controllable "
                             "but forward pass fails.")
                us[i] = np.nan
                xs[i + 1] = np.nan
                v_vec[i] = np.nan
            else:
                us[i] = optim_res[0]
                # The below function min( , max( ,)) ensure that the state x_{i+1} is controllable.
                # While this is ensured theoretically by the existence of the controllable sets,
                # numerical errors might violate this condition.
                xs[i + 1] = min(K[i + 1, 1], max(K[i + 1, 0], xs[i] + 2 * deltas[i] * us[i] - SMALL))
                v_vec[i] = optim_res[2:]
            logger.debug("[Forward pass] u_{:d} = {:f}, x_{:d} = {:f}".format(i, us[i], i + 1, xs[i + 1]))
        self.solver_wrapper.close_solver()
        # compute the slowest parametrization
        xs_slow = np.zeros(N + 1)
        xs_slow[0] = x_start
        us_slow = np.zeros(N)
        v_vec_slow = np.zeros((N, self.solver_wrapper.get_no_vars() - 2))

        self.solver_wrapper.setup_solver()
        for i in range(self._N):
            optim_res = self._forward_step(i, xs_slow[i], K[i + 1], slow=True)
            if np.isnan(optim_res[0]):
                logger.fatal("A numerical error occurs: The instance is controllable "
                             "but forward pass fails.")
                us_slow[i] = np.nan
                xs_slow[i + 1] = np.nan
                v_vec_slow[i] = np.nan
            else:
                us_slow[i] = optim_res[0]
                # The below function min( , max( ,)) ensure that the state x_{i+1} is controllable.
                # While this is ensured theoretically by the existence of the controllable sets,
                # numerical errors might violate this condition.
                xs_slow[i + 1] = min(K[i + 1, 1], max(K[i + 1, 0], xs_slow[i] + 2 * deltas[i] * us_slow[i] - SMALL))
                v_vec_slow[i] = optim_res[2:]
            logger.debug("[Forward pass] u_{:d} = {:f}, x_{:d} = {:f}".format(i, us_slow[i], i + 1, xs_slow[i + 1]))
        self.solver_wrapper.close_solver()

        # desired parametrization xs_desired = alpha * xs + (1 - alpha) / xs_slow
        duration = _compute_duration(xs, deltas)
        duration_slow = _compute_duration(xs_slow, deltas)
        if duration > self.desired_duration:
            logger.warn("Desired duration {:f} seconds is not achievable."
                        " Returning the fastest parameterization with duration {:f} seconds"
                        "".format(self.desired_duration, duration))
            alpha = 1.0
        elif duration_slow < self.desired_duration:
            logger.warn("Desired duration {:f} seconds is not achievable."
                        " Returning the slowest parameterization with duration {:f} seconds"
                        "".format(self.desired_duration, duration))
            alpha = .0
        else:
            logger.info("Desired duration {:f} sec is achievable. Continue computing.".format(self.desired_duration))
            alpha_low = 1.0  # here, low means a lower duration, and faster speed
            alpha_high = 0.0
            diff = 10
            it = 0
            while diff > atol:
                it += 1
                alpha = 0.5 * (alpha_low + alpha_high)
                xs_alpha = alpha * xs + (1 - alpha) * xs_slow
                duration_alpha = _compute_duration(xs_alpha, deltas)
                if duration_alpha < self.desired_duration:
                    alpha_low = alpha
                    diff = self.desired_duration - duration_alpha
                else:  # duration_alpha >= self.desired_duration
                    alpha_high = alpha
                    diff = duration_alpha - self.desired_duration
                logger.debug("it: {:d}. search range: [{:}, {:}]".format(
                    it, alpha_low, alpha_high))
        xs_alpha = alpha * xs + (1 - alpha) * xs_slow
        us_alpha = alpha * us + (1 - alpha) * us_slow
        v_vec_alpha = alpha * v_vec + (1 - alpha) * v_vec_slow

        sd_vec = np.sqrt(xs_alpha)
        sdd_vec = np.copy(us_alpha)
        self.problem_data.sd_vec = sd_vec
        self.problem_data.sdd_vec = sdd_vec
        if np.isnan(sd_vec).any():
            self.problem_data.return_code = algo.ParameterizationReturnCode.ErrUnknown
        else:
            self.problem_data.return_code = algo.ParameterizationReturnCode.Ok
        if return_data:
            return sdd_vec, sd_vec, v_vec_alpha, K
        else:
            return sdd_vec, sd_vec, v_vec_alpha

    def _forward_step(self, i, x, K_next, slow=False):
        """ Compute the highest possible path velocity that is controllable.

        Parameters
        ----------
        i: int
            Current stage index
        x: float
            The squared velocity at the current stage.
        K_next: list
            The lower and upper bounds of the set of controllable squared velocities
            in the next stage.
        eps: float, optional
            A numerical constant to avoid propagating numerical errors.
        slow: bool, optional

        Returns
        -------
        optim_var: array
            Optimal variable, which has this format (u, x, v).
            If this step fails, `optim_var` contains only nans.
        """
        # Immediate return
        if None in K_next or i < 0 or i > self._N or np.isnan(x) or x is None:
            return np.array([np.nan, np.nan])

        nV = self.solver_wrapper.get_no_vars()
        g_upper = np.zeros(nV)
        if not slow:
            g_upper[1] = - 1
            g_upper[0] = - 2 * self.solver_wrapper.get_deltas()[i]
        else:
            g_upper[1] = 1
            g_upper[0] = 2 * self.solver_wrapper.get_deltas()[i]

        # Account for propagating numerical errors
        K_next_max = K_next[1]
        K_next_min = K_next[0]

        optim_var = self.solver_wrapper.solve_stagewise_optim(
            i, None, g_upper, x, x, K_next_min, K_next_max)
        return optim_var



import numpy as np
from .constants import JVEL_MAXSD
cimport numpy as np
DTYPE = np.float64
ctypedef np.float64_t FLOAT_t

cdef inline np.float64_t float64_max(
    FLOAT_t a, FLOAT_t b): return a if a >= b else b
cdef inline np.float64_t float64_min(
    FLOAT_t a, FLOAT_t b): return a if a <= b else b
cdef inline np.float64_t float64_abs(
    FLOAT_t a): return a if a > 0 else - a

cdef float MAXSD = JVEL_MAXSD  # Maximum allowable path velocity for velocity constraint

cpdef _create_velocity_constraint(np.ndarray[double, ndim=2] qs,
                                  np.ndarray[double, ndim=2] vlim):
    """ Evaluate coefficient matrices for velocity constraints.

    A maximum allowable value for the path velocity is defined with
    `MAXSD`. This constant results in a lower bound on trajectory
    duration obtain by toppra.

    Args:
    ----
    qs: ndarray
        Path derivatives at each grid point.
    vlim: ndarray
        Velocity bounds.
    
    Returns:
    --------
    a: ndarray
    b: ndarray
    c: ndarray
        Coefficient matrices.
    """
    cdef int N = qs.shape[0] - 1
    cdef int dof = qs.shape[1]
    cdef int i, k
    cdef float sdmin, sdmax
    # Evaluate sdmin, sdmax at each steps and fill the matrices.
    cdef np.ndarray[np.float64_t, ndim=2] a = np.zeros((N + 1, 2), dtype=float)
    cdef np.ndarray[np.float64_t, ndim=2] b = np.ones((N + 1, 2), dtype=float)
    cdef np.ndarray[np.float64_t, ndim=2] c = np.zeros((N + 1, 2), dtype=float)
    b[:, 1] = -1
    for i in range(N + 1):
        sdmin = - MAXSD
        sdmax = MAXSD
        for k in range(dof):
            if qs[i, k] > 0:
                sdmax = float64_min(vlim[k, 1] / qs[i, k], sdmax)
                sdmin = float64_max(vlim[k, 0] / qs[i, k], sdmin)
            elif qs[i, k] < 0:
                sdmax = float64_min(vlim[k, 0] / qs[i, k], sdmax)
                sdmin = float64_max(vlim[k, 1] / qs[i, k], sdmin)
        c[i, 0] = - sdmax**2
        c[i, 1] = float64_max(sdmin, 0.)**2
    return a, b, c

cpdef _create_velocity_constraint_varying(np.ndarray[double, ndim=2] qs,
                                          np.ndarray[double, ndim=3] vlim_grid):
    """ Evaluate coefficient matrices for velocity constraints.

    Args:
    ----
    qs: (N,) ndarray
        Path derivatives at each grid point.
    vlim_grid: (N, dof, 2) ndarray
        Velocity bounds at each grid point.
    
    Returns:
    --------
    a: ndarray
    b: ndarray
    c: ndarray
        Coefficient matrices.
    """
    cdef int N = qs.shape[0] - 1
    cdef int dof = qs.shape[1]
    cdef int i, k
    cdef float sdmin, sdmax
    # Evaluate sdmin, sdmax at each steps and fill the matrices.
    cdef np.ndarray[np.float64_t, ndim=2] a = np.zeros((N + 1, 2), dtype=float)
    cdef np.ndarray[np.float64_t, ndim=2] b = np.ones((N + 1, 2), dtype=float)
    cdef np.ndarray[np.float64_t, ndim=2] c = np.zeros((N + 1, 2), dtype=float)
    b[:, 1] = -1
    for i in range(N + 1):
        sdmin = - MAXSD
        sdmax = MAXSD
        for k in range(dof):
            if qs[i, k] > 0:
                sdmax = float64_min(vlim_grid[i, k, 1] / qs[i, k], sdmax)
                sdmin = float64_max(vlim_grid[i, k, 0] / qs[i, k], sdmin)
            elif qs[i, k] < 0:
                sdmax = float64_min(vlim_grid[i, k, 0] / qs[i, k], sdmax)
                sdmin = float64_max(vlim_grid[i, k, 1] / qs[i, k], sdmin)
        c[i, 0] = - sdmax**2
        c[i, 1] = float64_max(sdmin, 0.)**2
    return a, b, c




from .solverwrapper import SolverWrapper, check_solver_availability
from ..constraint import ConstraintType
from ..constants import INFTY, ECOS_MAXX, ECOS_INFTY
import logging
import numpy as np
import scipy.sparse

try:
    import ecos
except ImportError:
    pass

logger = logging.getLogger(__name__)


class ecosWrapper(SolverWrapper):
    """A solver wrapper that handles linear and conic-quadratic constraints using ECOS.

    :class:`ecosWrapper` and :class:`cvxpyWrapper` are the only
    wrappers that can handle conic-quadratic constraints, which are
    necessary to compute robust path parameterization.

    Notes
    -----
    To reduce numerical-related issues, ECOS_MAXX is used to regulate
    the magnitude of the solution.

    ECOS is not very well implemented. There are many cases in which
    the solver fails simply because there is a very large bound
    (>1e6). Because of this, the test suites included in toppra do not
    include many tests for ecos.


    Attributes
    ----------
    constraints : list of `Constraint`
        Constraints on the robot system.
    path : Interpolator
        The geometric path to be time-parametrized.
    path_discretization: array
        The discretization grid use to discretize the geometric path.

    """

    def __init__(self, constraint_list, path, path_discretization):
        super(ecosWrapper, self).__init__(constraint_list, path, path_discretization)
        if not check_solver_availability('ecos'):
            raise RuntimeError(
                "Unable to start ecos wrapper because ECOS solver is not installed."
            )
        # NOTE: Currently receive params in dense form.
        self._linear_idx = []
        self._conic_idx = []
        # Linear_dim + conic_dim are the number of rows of matrix G and h
        # conic_dim is divisible by 4
        self._linear_dim = 0
        self._conic_dim = 0
        logger.debug("Registering constraints")
        for i, constraint in enumerate(constraint_list):
            logger.debug("Constraint [{:d}] \n {:}".format(i, constraint.__repr__()))
            _type = constraint.get_constraint_type()
            if _type == ConstraintType.CanonicalLinear:
                self._linear_idx.append(i)
                if self.params[i][3] is not None:  # Check F is not None
                    if constraint.identical:
                        self._linear_dim += self.params[i][3].shape[0]
                    else:
                        self._linear_dim += self.params[i][3].shape[1]
                if self.params[i][5] is not None:  # Check ubound
                    self._linear_dim += 2
                if self.params[i][6] is not None:  # Check xbound
                    self._linear_dim += 2
            elif _type == ConstraintType.CanonicalConic:
                self._conic_idx.append(i)
                self._conic_dim += 4 * self.params[i][0].shape[1]
            else:
                raise NotImplementedError(
                    "Constraint type {:} not implemented".format(_type)
                )
        assert self._conic_dim % 4 == 0
        logger.debug("Indices of linear constraints: {:}".format(self._linear_idx))
        logger.debug("Indices of conic constraints : {:}".format(self._conic_idx))
        logger.debug(
            "Nb of row for linear constraints: {:d} rows".format(self._linear_dim)
        )
        logger.debug(
            "Nb of row for conic constraints : {:d} rows".format(self._conic_dim)
        )

    def solve_stagewise_optim(self, i, H, g, x_min, x_max, x_next_min, x_next_max):
        assert i <= self.N and 0 <= i
        assert H is None or np.allclose(H, np.zeros(2))

        # Total number of rows in matrix G, h. Descriptions of terms
        # 1) x_min <= x <= x_max
        # 2) xnext_min <= x + 2 ds u <= xnext_max
        # 3) linear constraints
        # 4) conic constraints
        if i < self.N:
            nrow = 2 + 2 + self._linear_dim + self._conic_dim
            dims = {"l": 2 + 2 + self._linear_dim, "q": [4] * (self._conic_dim // 4)}
        else:
            nrow = 2 + self._linear_dim + self._conic_dim
            dims = {"l": 2 + self._linear_dim, "q": [4] * (self._conic_dim // 4)}

        G_lil = scipy.sparse.lil_matrix((nrow, 2))
        h = np.zeros(nrow)
        # Fill G and h
        currow = 0
        ## Fill 1)
        G_lil[currow : currow + 2, 1] = [[-1], [1]]
        if not np.isnan(x_min):
            h[currow] = -x_min
        else:
            h[currow] = ECOS_INFTY
        currow += 1
        if not np.isnan(x_max):
            h[currow] = x_max
        else:
            h[currow] = ECOS_INFTY
        currow += 1
        ## Fill 2)
        if i < self.N:
            delta = self.get_deltas()[i]
            G_lil[currow, :] = [[-2 * delta, -1]]
            if not np.isnan(x_next_min):
                h[currow] = -x_next_min
            else:
                h[currow] = ECOS_INFTY
            currow += 1
            G_lil[currow, :] = [[2 * delta, 1]]
            if not np.isnan(x_next_max):
                h[currow] = x_next_max
            else:
                h[currow] = ECOS_INFTY
            currow += 1
        ## Fill 3)
        for k in self._linear_idx:
            _a, _b, _c, _F, _h, _ubound, _xbound = self.params[k]

            if _a is not None:
                if self.constraints[k].identical:
                    nb_cnst = _F.shape[0]
                    G_lil[currow : currow + nb_cnst, 0] = np.dot(_F, _a[i]).reshape(
                        -1, 1
                    )
                    G_lil[currow : currow + nb_cnst, 1] = np.dot(_F, _b[i]).reshape(
                        -1, 1
                    )
                    h[currow : currow + nb_cnst] = _h - np.dot(_F, _c[i])
                    currow += nb_cnst
                else:
                    nb_cnst = _F.shape[1]
                    G_lil[currow : currow + nb_cnst, 0] = np.dot(_F[i], _a[i]).reshape(
                        -1, 1
                    )
                    G_lil[currow : currow + nb_cnst, 1] = np.dot(_F[i], _b[i]).reshape(
                        -1, 1
                    )
                    h[currow : currow + nb_cnst] = _h[i] - np.dot(_F[i], _c[i])
                    currow += nb_cnst

            if _ubound is not None:
                G_lil[currow, 0] = 1
                G_lil[currow + 1, 0] = -1
                h[currow : currow + 2] = [_ubound[i, 1], -_ubound[i, 0]]
                currow += 2

            if _xbound is not None:
                G_lil[currow, 1] = 1
                G_lil[currow + 1, 1] = -1
                h[currow : currow + 2] = [min(ECOS_MAXX, _xbound[i, 1]), -_xbound[i, 0]]
                currow += 2
        ## Fill 4)
        for k in self._conic_idx:
            _a, _b, _c, _P, _, _ = self.params[k]

            # NOTE: Here the following arrangement is used.
            # G_i = [  a_ij        b_ij  ]
            #       [ -P_ij^T[:, :2]     ]
            # h_i = [- c_ij         ]
            #       [- P_ij^T[:, 2] ]
            for j in range(_a.shape[1]):
                G_lil[currow, :] = [_a[i, j], _b[i, j]]
                G_lil[currow + 1 : currow + 4, :] = _P[i, j].T[:, :2]
                h[currow] = -_c[i, j]
                h[currow + 1 : currow + 4] = _P[i, j].T[:, 2]
                currow += 4

        # Fill
        G = scipy.sparse.csc_matrix(G_lil)
        result = ecos.solve(g, G, h, dims, verbose=False)
        accepted_infos = ["Optimal solution found", "Close to optimal solution found"]
        if result["info"]["infostring"] in accepted_infos:
            success = True
        else:
            success = False
            logger.warning(
                "Optimization fails. Result dictionary: \n {:}".format(result)
            )

        ux_opt = np.zeros(2)
        if success:
            ux_opt = result["x"]
        else:
            ux_opt[:] = np.nan
        return ux_opt



"""This module implements the general Second-Order constraints."""
import logging
import numpy as np

from .linear_constraint import LinearConstraint, canlinear_colloc_to_interpolate
from .constraint import DiscretizationType

logger = logging.getLogger(__name__)


class SecondOrderConstraint(LinearConstraint):
    """This class implements the linear Second-Order constraint.

    Conventionally, a :class:`SecondOrderConstraint` is given by the
    following formula:

    .. math::
        A(\mathbf{q}) \ddot {\mathbf{q}} + \dot
        {\mathbf{q}}^\\top B(\mathbf{q}) \dot {\mathbf{q}} + C(\mathbf{q}) = w,

    where w is a vector that satisfies the polyhedral constraint:

    .. math::
        F(\mathbf{q}) w \\leq g(\mathbf{q}).

    Take the example of a robot torque bound, the functions :math:`A,
    B, C` represent respectively the inertial, Corriolis and
    gravitational terms of the robot's rigid body dynamics.

    We can evaluate a constraint on a given geometric path
    :math:`\mathbf{p}(s)` using the following equations, which are
    obtained by direct substitution:

    .. math::

        A(\mathbf{q}) \mathbf{p}'(s) \ddot s + [A(\mathbf{q}) \mathbf{p}''(s) + \mathbf{p}'(s)^\\top B(\mathbf{q})
        \mathbf{p}'(s)] \dot s^2 + C(\mathbf{q}) = w, \\\\
        \mathbf{a}(s) \ddot s + \mathbf{b}(s) \dot s ^2 + \mathbf{c}(s) = w.

    where :math:`\mathbf{p}', \mathbf{p}''` denote respectively the
    first and second derivatives of the path. It is important to
    understand that the vector functions :math:`\mathbf a, \mathbf b,
    \mathbf c` are what `toppra` needs to solve for path
    parametrizations.

    To evaluate these coefficients :math:`\mathbf a(s), \mathbf b(s),
    \mathbf c(s)`, fortunately, it is not necessary to have the
    functions :math:`A, B, C` explicitly. Rather, it is only required
    to have the sum of the these 3 functions--the so-called inverse
    dynamic function:

    .. math::
        \mathrm{inverse\_dyn}(\mathbf q, \dot{\mathbf q}, \ddot{\mathbf q}) :=
        A(\mathbf{q}) \ddot {\mathbf{q}} + \dot {\mathbf{q}}^\\top B(\mathbf{q}) \dot {\mathbf{q}} + C(\mathbf{q})

    In some cases, one might have terms that depends purely on the
    path:

    .. math::
        \mathbf{a}(s) \ddot s + \mathbf{b}(s) \dot s ^2 + \mathbf{c}(s) + \mathcal{C}(\mathbf p, s)= w.

    an example is the joint friction. This term is referred to as
    `custom_term` in the initializing arguments of
    :class:`SecondOrderConstraint`.

    It is interesting to note that we can actually use a more general
    form of the above equations, hence covering a wider class of
    constraints. In particular, one can replace :math:`A(\mathbf{q}),
    B(\mathbf{q}), C(\mathbf{q}), F(\mathbf{q}), g(\mathbf{q})` with
    :math:`A(\mathbf{q}, s), B(\mathbf{q}, s), C(\mathbf{q}, s),
    F(\mathbf{q}, s), g(\mathbf{q}, s)`. This form, however, is not
    implemented in `toppra`.

    """

    def __init__(self, inv_dyn, constraint_F, constraint_g, dof, custom_term=None, discretization_scheme=1):
        """Initialize the constraint.

        Parameters
        ----------
        inv_dyn: (np.ndarray, np.ndarray, np.ndarray) -> np.ndarray
            The "inverse dynamics" function that receives joint
            position, velocity, acceleration and path position as inputs and ouputs
            the constrained vector :math:`\mathbf w`. See above for more details.
        constraint_F: (np.ndarray) -> np.ndarray
            The constraint coefficient function :math:`\mathbf
            F(\mathbf q, s)`. See above for more details.
        constraint_g: (np.ndarray) -> np.ndarray
            The constraint coefficient function :math:`\mathbf
            g(\mathbf q, s)`. See above for more details.
        dof: int
            The dimension of the joint position.
        custom_term: (:class:`Interpolator`, float) -> np.ndarray
            This function receives as input a geometric path and a
            float path position, then returns an additive term. See
            the above note for more details.

        discretization_scheme: DiscretizationType
            Type of discretization.

        """
        super(SecondOrderConstraint, self).__init__()
        self.set_discretization_type(discretization_scheme)
        self.inv_dyn = inv_dyn
        self.constraint_F = constraint_F
        self.constraint_g = constraint_g
        self.dof = dof
        self.custom_term = custom_term
        self._format_string = "    Kind: Generalized Second-order constraint\n"
        self._format_string = "    Dimension:\n"
        self._format_string += "        F in R^({:d}, {:d})\n".format(
            *constraint_F(np.zeros(dof)).shape)

    @classmethod
    def joint_torque_constraint(cls, inv_dyn, taulim, joint_friction,
                                **kwargs):
        """Initialize a Joint Torque constraint.

        Parameters
        ----------
        inv_dyn: (np.ndarray, np.ndarray, np.ndarray) -> np.ndarray
            Inverse dynamic function of the robot.
        taulim: np.ndarray
            Shape (N, 2). The i-th element contains the minimum and
            maximum joint torque limits respectively.
        joint_friction: np.ndarray
            Shape (N,). The i-th element contains the dry friction (torque)
            in the i-th joint.

        """
        dof = np.shape(taulim)[0]
        stacked_eyes = np.vstack((np.eye(dof), -np.eye(dof)))
        g_aug = np.zeros(2 * dof)
        g_aug[:dof] = taulim[:, 1]
        g_aug[dof:] = -taulim[:, 0]
        constraint_F = lambda _: stacked_eyes
        constraint_g = lambda _: g_aug
        custom_term = lambda path, s: np.sign(path(s, 1)) * joint_friction
        return SecondOrderConstraint(inv_dyn, constraint_F, constraint_g, dof, custom_term,
                                     **kwargs)

    def compute_constraint_params(self, path, gridpoints):
        if path.dof != self.dof:
            raise ValueError(
                "Wrong dimension: constraint dof ({:d}) not equal to path dof ({:d})"
                .format(self.dof, path.dof))
        v_zero = np.zeros(path.dof)
        p_vec = path(gridpoints)
        ps_vec = path(gridpoints, 1)
        pss_vec = path(gridpoints, 2)

        F_vec = np.array(list(map(self.constraint_F, p_vec)))
        g_vec = np.array(list(map(self.constraint_g, p_vec)))
        c_vec = np.array([self.inv_dyn(_p, v_zero, v_zero) for _p in p_vec])

        a_vec = np.array(
            [self.inv_dyn(_p, v_zero, _ps)
             for _p, _ps in zip(p_vec, ps_vec)]) - c_vec
        b_vec = np.array([
            self.inv_dyn(_p, _ps, pss_)
            for _p, _ps, pss_ in zip(p_vec, ps_vec, pss_vec)
        ]) - c_vec
        if self.custom_term is not None:
            for i, _ in enumerate(gridpoints):
                c_vec[i] = c_vec[i] + self.custom_term(path, gridpoints[i])

        if self.discretization_type == DiscretizationType.Collocation:
            return a_vec, b_vec, c_vec, F_vec, g_vec, None, None
        if self.discretization_type == DiscretizationType.Interpolation:
            return canlinear_colloc_to_interpolate(a_vec, b_vec, c_vec, F_vec,
                                                   g_vec, None, None,
                                                   gridpoints)
        raise NotImplementedError("Other form of discretization not supported!")



"""
TOPP-RA
=======
Reachability-Analysis-based Time-Optimal Path Parametrization.

This package produces routines for creation and handling path constraints
using the algorithm `TOPP-RA`.
"""
import logging

# core modules
from .interpolator import RaveTrajectoryWrapper, SplineInterpolator,\
    UnivariateSplineInterpolator, PolynomialPath
from .simplepath import SimplePath
from .parametrizer import ParametrizeConstAccel, ParametrizeSpline
from . import constraint
from . import algorithm
from . import solverwrapper

# utility
from .utils import smooth_singularities, setup_logging
from .planning_utils import retime_active_joints_kinematics,\
    create_rave_torque_path_constraint

# set nullhandler by default
logging.getLogger('toppra').addHandler(logging.NullHandler())
